import pydlm
import unittest

class testTrend(unittest.TestCase):

    def setUp(self):
        self.DEGREE = 3
        
    def testInitialization(self):
        newTrend = pydlm.modeler.trends.trend(self.DEGREE)
        newTrend.checkDimensions()

unittest.main()



"""
=========================================================================

Code for the dynamic component

=========================================================================

This piece of code provide one building block for the dynamic linear model.
It decribes a dynamic component in the time series data. It basically allows
user to supply covariate or controlled variable into the dlm, 
and the coefficients of the features will be trained as the latent states. 
Examples are holiday indicators, other observed variables and so on.

The name dynamic means that the features are changing over time.

"""
import numpy as np
from .component import component
import pydlm.base.tools as tl

# create trend component
# We create the trend using the component class

class dynamic(component):
    """ The dynamic component that allows user to add controlled variables, 
    providing one building block for the dynamic linear model.
    It decribes a dynamic component in the time series data. It basically allows
    user to supply covariate or controlled variable into the dlm, 
    and the coefficients of the features will be trained as the latent states. 
    Examples are holiday indicators, other observed variables and so on.
 
    Examples:
        >>>  # create a dynamic component:
        >>> features = [[1.0, 2.0] for i in range(10)]
        >>> ctrend = dynamic(features = features, name = 'random', discount = 0.99)
        >>>  # change the ctrend to have covariance with diagonals are 2 and state 1
        >>> ctrend.createCovPrior(cov = 2)
        >>> ctrend.createMeanPrior(mean = 1)
   
    Attributes:
        d: the dimension of the features (number of latent states)
        n: the number of observation
        componentType: the type of the component, in this case, 'dynamic'
        name: the name of the trend component, to be supplied by user
              used in modeling and result extraction
        discount: the discount factor for this component. Details please refer
                  to the @kalmanFilter
        evaluation: the evaluation matrix for this component
        transition: the transition matrix for this component
        covPrior: the prior guess of the covariance matrix of the latent states
        meanPrior: the prior guess of the latent states
    
    """
    def __init__(self, features = None, discount = 0.99, name = 'dynamic'):
        self.n = len(features)
        self.d = len(features[0])
        self.features = tl.duplicateList(features)
        self.componentType = 'dynamic'
        self.name = name
        self.discount = np.ones(self.d) * discount
        
        # Initialize all basic quantities
        self.evaluation = None
        self.transition = None
        self.covPrior = None
        self.meanPrior = None

        # create all basic quantities
        self.createEvaluation(0)
        self.createTransition()
        self.createCovPrior()
        self.createMeanPrior()

        # record current step in case of lost
        self.step = 0

    def createEvaluation(self, step):
        """ The evaluation matrix for the dynamic component change over time.
        It equals to the value of the features or the controlled variables at a 
        given date

        """
        self.evaluation = np.matrix([self.features[step]])

    def createTransition(self):
        """ Create the transition matrix.

        For the dynamic component, the transition matrix is just the identity matrix

        """
        self.transition = np.matrix(np.eye(self.d))
        
    def createCovPrior(self, cov = None, scale = 1):
        """ Create the prior covariance matrix for the latent states

        """
        if cov is None:
            self.covPrior = np.matrix(np.eye(self.d)) * scale
        else:
            self.covPrior = cov * scale

    def createMeanPrior(self, mean = None, scale = 1):
        """ Create the prior latent state
        
        """
        if mean is None:
            self.meanPrior = np.matrix(np.zeros((self.d, 1))) * scale
        else:
            self.meanPrior = mean * scale

    def checkDimensions(self):
        """ if user supplies their own covPrior and meanPrior, this can
        be used to check if the dimension matches

        """
        tl.checker.checkVectorDimension(self.meanPrior, self.covPrior)
        print('The dimesnion looks good!')

    def updateEvaluation(self, step):
        """ update the evaluation matrix to a specific date
        This function is used when fitting the forward filter and backward smoother
        in need of updating the correct evaluation matrix

        """
        if step < self.n:
            self.evaluation = np.matrix([self.features[step]])
            self.step = step
        else:
            raise NameError('The step is out of range')

    def appendNewData(self, newData):
        """ For updating feature matrix when new data is added.

        Args:
            newData: is a list of list. The inner list is the feature vector. The outer
                     list may contain multiple feature vectors.

        """
        self.features.extend(tl.duplicateList(newData))
        self.n = len(self.features)

    def popout(self, date):
        """ For deleting the feature data of a specific date.
        
        Args:
            date: the index of which to be deleted.

        """
        self.features.pop(date)
        self.n -= 1



"""
=========================================================================

Code for the trend component

=========================================================================

This piece of code provide one building block for the dynamic linear model.
It decribes a latent polynomial trending in the time series data.

"""
import numpy as np
from .component import component
import pydlm.base.tools as tl

# create trend component
# We create the trend using the component class

class trend(component):
    """ The trend component that features the polynomial trending, 
    providing one building block for the dynamic linear model.
    It decribes a latent polynomial trending in the time series data.

    Examples:
        >>>  # create a constant trend
        >>> ctrend = trend(degree = 1, name = 'Const', discount = 0.99)
        >>>  # change the ctrend to have covariance with diagonals are 2 and state 1
        >>> ctrend.createCovPrior(cov = 2)
        >>> ctrend.createMeanPrior(mean = 1)
    
    Attributes:
        d: the degree of the polynomial trend
        componentType: the type of the component, in this case, 'trend'
        name: the name of the trend component, to be supplied by user
              used in modeling and result extraction
        discount: the discount factor for this component. Details please refer
                  to the @kalmanFilter
        evaluation: the evaluation matrix for this component
        transition: the transition matrix for this component
        covPrior: the prior guess of the covariance matrix of the latent states
        meanPrior: the prior guess of the latent states
    
    """
    
    def __init__(self, degree = 1, discount = 0.99, name = 'trend'):
        if degree <= 0:
            raise NameError('degree has to be positive')
        self.d = degree
        self.name = name
        self.componentType = 'trend'
        self.discount = np.ones(self.d) * discount
        
        # Initialize all basic quantities
        self.evaluation = None
        self.transition = None
        self.covPrior = None
        self.meanPrior = None

        # create all basic quantities
        self.createEvaluation()
        self.createTransition()
        self.createCovPrior()
        self.createMeanPrior()

    def createEvaluation(self):
        """ Create the evaluation matrix

        """
        self.evaluation = np.matrix(np.zeros((1, self.d)))
        self.evaluation[0, 0] = 1

    def createTransition(self):
        """Create the transition matrix

        According Hurrison and West (1999), the transition matrix of trend takes
        a form of \n

        [[1 1 1 1],\n
        [0 1 1 1],\n
        [0 0 1 1],\n
        [0 0 0 1]]

        """
        self.transition = np.matrix(np.zeros((self.d, self.d)))
        self.transition[np.triu_indices(self.d)] = 1
        
    def createCovPrior(self, cov = 1):
        """Create the prior covariance matrix for the latent states.

        """
        self.covPrior = np.matrix(np.eye(self.d)) * cov

    def createMeanPrior(self, mean = 0):
        """ Create the prior latent state
    
        """
        self.meanPrior = np.matrix(np.ones((self.d, 1))) * mean

    def checkDimensions(self):
        """ if user supplies their own covPrior and meanPrior, this can 
        be used to check if the dimension matches

        """
        tl.checker.checkVectorDimension(self.meanPrior, self.covPrior)
        print('The dimesnion looks good!')



"""
=========================================================================

Code for the seasonality component

=========================================================================

This piece of code provide one building block for the dynamic linear model.
It decribes a latent seasonality trending in the time series data. The user
can use this class to construct any periodicy component to the model, for
instance, the hourly, weekly or monthly behavior. Different from the Fourier
series, the seasonality components are nonparametric, i.e., there is no sin
or cos relationship between each state. They can be arbitrarily valued.

"""
import numpy as np
from .component import component
import pydlm.base.tools as tl

# create seasonality component
# We create the seasonality using the component class

class seasonality(component):
    """The seasonality component that features the periodicity behavior,
    providing one building block for the dynamic linear model.
    It decribes a latent seasonality trending in the time series data. The user
    can use this class to construct any periodicy component to the model, for
    instance, the hourly, weekly or monthly behavior. Different from the Fourier
    series, the seasonality components are nonparametric, i.e., there is no sin
    or cos relationship between each state. They can be arbitrarily valued.
 
    Examples:
        >>>  # create a 7-day seasonality:
        >>> ctrend = seasonality(period = 7, name = 'weekly', discount = 0.99)
        >>>  # change the ctrend to have covariance with diagonals are 2 and state 1
        >>> ctrend.createCovPrior(cov = 2)
        >>> ctrend.createMeanPrior(mean = 1)
        >>> ctrend.freeForm()
   
    Attributes:
        d: the period of the seasonality
        componentType: the type of the component, in this case, 'seasonality'
        name: the name of the seasonality component, to be supplied by user
              used in modeling and result extraction
        discount: the discount factor for this component. Details please refer
                  to the @kalmanFilter
        evaluation: the evaluation matrix for this component
        transition: the transition matrix for this component
        covPrior: the prior guess of the covariance matrix of the latent states
        meanPrior: the prior guess of the latent states

    
    """
    def __init__(self, period = 7, discount = 0.99, name = 'seasonality'):
        if period <= 1:
            raise NameError('Period has to be greater than 1.')
        self.d = period
        self.componentType = 'seasonality'
        self.name = name
        self.discount = np.ones(self.d) * discount
        
        # Initialize all basic quantities
        self.evaluation = None
        self.transition = None
        self.covPrior = None
        self.meanPrior = None

        # create all basic quantities
        self.createEvaluation()
        self.createTransition()
        self.createCovPrior()
        self.createMeanPrior()

        # create form free seasonality component
        self.freeForm()

    def createEvaluation(self):
        """ Create the evaluation matrix
        
        """
        self.evaluation = np.matrix(np.zeros((1, self.d)))
        self.evaluation[0, 0] = 1

    # The transition matrix takes special form as
    # G = [0 1 0]
    #     [0 0 1]
    #     [1 0 0]
    # So everyt time, when G * G, we rotate the vector once, which results
    # in the seasonality performance
    def createTransition(self):
        """ Create the transition matrix.
        
        According to Hurrison and West (1999), the transition matrix of seasonality
        takes a form of\n

        [[0 1 0 0],\n
        [0 0 1 0],\n
        [0 0 0 1],\n
        [1 0 0 0]]

        """
        self.transition = np.matrix(np.diag(np.ones(self.d - 1), 1))
        self.transition[self.d - 1, 0] = 1
        
    def createCovPrior(self, cov = 1):
        """Create the prior covariance matrix for the latent states.
        
        """
        self.covPrior = np.matrix(np.eye(self.d)) * cov

    def createMeanPrior(self, mean = 0):
        """ Create the prior latent state
    
        """
        self.meanPrior = np.matrix(np.ones((self.d, 1))) * mean

    # Form free seasonality component ensures that sum(mean) = 0
    # We use the formular from "Bayesian forecasting and dynamic linear models"
    # Page 242
    def freeForm(self):
        """ The technique used in Hurrison and West (1999). After calling this method,
        The latent states sum up to 0 and the covariance matrix is degenerate to have
        rank d - 1, so that the sum of the latent states will never change when the
        system evolves

        """
        if self.covPrior is None or self.meanPrior is None:
            raise NameError('freeForm can only be called after prior created.')
        else:
            u = np.sum(np.sum(self.covPrior, 0), 1)[0, 0]
            A = np.sum(self.covPrior, 1) / u
            self.meanPrior = self.meanPrior - A * np.sum(self.meanPrior, 0)[0, 0]
            self.covPrior = self.covPrior - np.dot(A, A.T) * u
        
    def checkDimensions(self):
        """ if user supplies their own covPrior and meanPrior, this can 
        be used to check if the dimension matches

        """
        tl.checker.checkVectorDimension(self.meanPrior, self.covPrior)
        print('The dimension looks good!')



"""
==================================================================================

Code for plot dlm results

==================================================================================

This module provide the plotting functionality for dlm

"""
import matplotlib.pyplot as plt
from pydlm.base.tools import getInterval

# =========================== plot for main data ============================


def plotInOneFigure(time, data, result, options):
    """
    Plot the dlm results in one figure

    Args:
        time: the time label
        data: the original data
        result: the fitted result from dlm class
        options: options for the plot, for details please refer to @dlm
    """
    # plot the original data
    plotData(time=time, data=data,
             showDataPoint=options.showDataPoint, color=options.dataColor,
             label='time series')

    # plot fitered results if needed
    if options.plotFilteredData:
        start = result.filteredSteps[0]
        end = result.filteredSteps[1] + 1
        plotData(time=time[start:end],
                 data=to1dArray(result.filteredObs[start:end]),
                 showDataPoint=options.showFittedPoint,
                 color=options.filteredColor,
                 label='filtered series')

        if options.showConfidenceInterval:
            upper, lower = getInterval(result.filteredObs[start:end],
                                       result.filteredObsVar[start:end],
                                       p=options.confidence)

            plotInterval(time=time[start:end],
                         upper=to1dArray(upper), lower=to1dArray(lower),
                         intervalType=options.intervalType,
                         color=options.filteredColor)

        # plot predicted results if needed
    if options.plotPredictedData:
        start = result.filteredSteps[0]
        end = result.filteredSteps[1] + 1
        plotData(time=time[start:end],
                 data=to1dArray(result.predictedObs),
                 showDataPoint=options.showFittedPoint,
                 color=options.predictedColor,
                 label='one-day prediction')

        if options.showConfidenceInterval:
            upper, lower = getInterval(result.predictedObs[start:end],
                                       result.filteredObsVar[start:end],
                                       p=options.confidence)

            plotInterval(time=time[start:end],
                         upper=to1dArray(upper), lower=to1dArray(lower),
                         intervalType=options.intervalType,
                         color=options.predictedColor)

    # plot smoothed results if needed
    if options.plotSmoothedData:
        start = result.smoothedSteps[0]
        end = result.smoothedSteps[1] + 1
        plotData(time=time[start:end],
                 data=to1dArray(result.smoothedObs),
                 showDataPoint=options.showFittedPoint,
                 color=options.smoothedColor,
                 label='smoothed series')

        if options.showConfidenceInterval:
            upper, lower = getInterval(result.smoothedObs[start:end],
                                       result.smoothedObsVar[start:end],
                                       p=options.confidence)

            plotInterval(time=time[start:end],
                         upper=to1dArray(upper), lower=to1dArray(lower),
                         intervalType=options.intervalType,
                         color=options.smoothedColor)

    plt.legend(loc='best', shadow=True)  # , fontsize = 'x-large')


def plotInMultipleFigure(time, data, result, options):
    """
    Plot the dlm results in multiple figure, each with one result and the
    original data

    Args:
        time: the time label
        data: the original data
        result: the fitted result from dlm class
        options: options for the plot, for details please refer to @dlm
    """
    # first compute how many plots are needed
    numOfPlots = options.plotFilteredData + options.plotPredictedData + \
                 options.plotSmoothedData
    size = (numOfPlots, 1)
    location = 1

    # plot all needed results
    # plot fitered results if needed
    if options.plotFilteredData:
        # the location
        subplot(size, location)

        # plot original data
        plotData(time=time, data=data,
                 showDataPoint=options.showDataPoint, color=options.dataColor,
                 label='time series')

        start = result.filteredSteps[0]
        end = result.filteredSteps[1] + 1
        if start < end:
            plotData(time=time[start:end],
                     data=to1dArray(result.filteredObs[start:end]),
                     showDataPoint=options.showFittedPoint,
                     color=options.filteredColor,
                     label='filtered series')

            if options.showConfidenceInterval:
                upper, lower = getInterval(result.filteredObs[start:end],
                                           result.filteredObsVar[start:end],
                                           p=options.confidence)

                plotInterval(time=time[start:end],
                             upper=to1dArray(upper), lower=to1dArray(lower),
                             intervalType=options.intervalType,
                             color=options.filteredColor)
        plt.legend(loc='best', shadow=True)  # , fontsize = 'x-large')
        location += 1

    # plot predicted results if needed
    if options.plotPredictedData:
        # the location
        subplot(size, location)

        # plot original data
        plotData(time=time, data=data,
                 showDataPoint=options.showDataPoint,
                 color=options.dataColor, label='time series')

        start = result.filteredSteps[0]
        end = result.filteredSteps[1] + 1
        if start < end:
            plotData(time=time[start:end],
                     data=to1dArray(result.predictedObs),
                     showDataPoint=options.showFittedPoint,
                     color=options.predictedColor,
                     label='one-day prediction')

            if options.showConfidenceInterval:
                upper, lower = getInterval(result.predictedObs[start:end],
                                           result.filteredObsVar[start:end],
                                           p=options.confidence)
                plotInterval(time=time[start:end],
                             upper=to1dArray(upper), lower=to1dArray(lower),
                             intervalType=options.intervalType,
                             color=options.predictedColor)
        plt.legend(loc='best', shadow=True)
        location += 1

    # plot smoothed results if needed
    if options.plotSmoothedData:
        # the location
        subplot(size, location)

        # plot original data
        plotData(time=time, data=data,
                 showDataPoint=options.showDataPoint,
                 color=options.dataColor, label='time series')

        start = result.smoothedSteps[0]
        end = result.smoothedSteps[1] + 1
        if start < end:
            plotData(time=time[start:end],
                     data=to1dArray(result.smoothedObs),
                     showDataPoint=options.showFittedPoint,
                     color=options.smoothedColor,
                     label='smoothed series')

            if options.showConfidenceInterval:
                upper, lower = getInterval(result.smoothedObs[start:end],
                                           result.smoothedObsVar[start:end],
                                           p=options.confidence)
                plotInterval(time=time[start:end],
                             upper=to1dArray(upper), lower=to1dArray(lower),
                             intervalType=options.intervalType,
                             color=options.smoothedColor)
        plt.legend(loc='best', shadow=True)

# ============================ plot for latents ============================


def plotLatentState(time, coordinates, result, options, name):
    """
    Plot the latent state for given coordinates

    Args:
        time: the time label
        coordinates: the coordinates to plot
        result: the fitted result from dlm class
        options: options for the plot, for details please refer to @dlm
        name: the name of the component
    """
    # first compute how many plots are needed
    numOfPlots = len(coordinates)
    size = (numOfPlots, 1)

    # plot all needed results
    for i, dim in enumerate(coordinates):
        subplot(size, i + 1)
        plotSingleState(time, dim, result, options)
        plt.title('Filter result for dimension ' + str(i) +
                  ' in component: ' + name)


def plotSingleState(time, dimension, result, options):
    """
    Plot a single coordinate in the latent state vector

    Args:
        time: the time label
        dimension: the coordinate of the plot
        result: the fitted result from dlm class
        options: options for the plot, for details please refer to @dlm
    """

    # plot fitered results if needed
    if options.plotFilteredData:
        start = result.filteredSteps[0]
        end = result.filteredSteps[1] + 1
        data = [item[dimension, 0] for item in result.filteredState[start:end]]
        var = [abs(item[dimension, dimension])
               for item in result.filteredCov[start:end]]

        plotData(time=time[start:end],
                 data=data,
                 showDataPoint=options.showFittedPoint,
                 color=options.filteredColor,
                 label='filtered state')

        if options.showConfidenceInterval:
            upper, lower = getInterval(data, var, p=options.confidence)

            plotInterval(time=time[start:end],
                         upper=upper, lower=lower,
                         intervalType=options.intervalType,
                         color=options.filteredColor)

    # plot predicted results if needed
    if options.plotPredictedData:
        start = result.filteredSteps[0]
        end = result.filteredSteps[1] + 1
        data = [item[dimension, 0]
                for item in result.predictedState[start:end]]
        var = [abs(item[dimension, dimension])
               for item in result.predictedCov[start:end]]

        plotData(time=time[start:end],
                 data=data,
                 showDataPoint=options.showFittedPoint,
                 color=options.predictedColor,
                 label='predicted state')

        if options.showConfidenceInterval:
            upper, lower = getInterval(data, var, p=options.confidence)

            plotInterval(time=time[start:end],
                         upper=upper, lower=lower,
                         intervalType=options.intervalType,
                         color=options.predictedColor)

    # plot smoothed results if needed
    if options.plotSmoothedData:
        start = result.smoothedSteps[0]
        end = result.smoothedSteps[1] + 1
        data = [item[dimension, 0] for item in result.smoothedState[start:end]]
        var = [abs(item[dimension, dimension])
               for item in result.smoothedCov[start:end]]

        plotData(time=time[start:end],
                 data=data,
                 showDataPoint=options.showFittedPoint,
                 color=options.smoothedColor,
                 label='smoothed state')

        if options.showConfidenceInterval:
            upper, lower = getInterval(data, var, p=options.confidence)

            plotInterval(time=time[start:end],
                         upper=upper, lower=lower,
                         intervalType=options.intervalType,
                         color=options.smoothedColor)

    plt.legend(loc='best', shadow=True)

# ============================ plot for component =============================


def plotComponent(time, data, result, options):
    """
    Plot the results of a single component in one figure

    Args:
        time: the time label
        data: a dictionary contains all information. The basic keys are
              data['filteredMean'] = ...
              data['filteredVar']  = ...
              data['predictedMean']= ...
              data['predictedVar'] = ...
              data['smoothedMean'] = ...
              data['smoothedVar']  = ...
              data['name']         = ...
              not all are needed, depending on the setting in options.
        result: the fitted result from dlm class
        options: options for the plot, for details please refer to @dlm
    """

    if options.separatePlot:
        numOfPlots = options.plotFilteredData + options.plotPredictedData + \
                     options.plotSmoothedData
        size = (numOfPlots, 1)
        location = 1

    # plot fitered results if needed
    if options.plotFilteredData:
        if options.separatePlot:
            subplot(size, location)
            location += 1

        start = result.filteredSteps[0]
        end = result.filteredSteps[1] + 1
        plotData(time=time[start:end],
                 data=data['filteredMean'],
                 showDataPoint=options.showFittedPoint,
                 color=options.filteredColor,
                 label='filtered ' + data['name'])

        if options.showConfidenceInterval:
            upper, lower = getInterval(data['filteredMean'],
                                       map(abs, data['filteredVar']),
                                       p=options.confidence)

            plotInterval(time=time[start:end],
                         upper=upper, lower=lower,
                         intervalType=options.intervalType,
                         color=options.filteredColor)

        plt.legend(loc='best', shadow=True)

    # plot predicted results if needed
    if options.plotPredictedData:
        if options.separatePlot:
            subplot(size, location)
            location += 1

        start = result.filteredSteps[0]
        end = result.filteredSteps[1] + 1
        plotData(time=time[start:end],
                 data=data['predictedMean'],
                 showDataPoint=options.showFittedPoint,
                 color=options.predictedColor,
                 label='one-step predict for ' + data['name'])

        if options.showConfidenceInterval:
            upper, lower = getInterval(data['predictedMean'],
                                       map(abs, data['predictedVar']),
                                       p=options.confidence)

            plotInterval(time=time[start:end],
                         upper=upper, lower=lower,
                         intervalType=options.intervalType,
                         color=options.predictedColor)

        plt.legend(loc='best', shadow=True)

    # plot smoothed results if needed
    if options.plotSmoothedData:
        if options.separatePlot:
            subplot(size, location)
            location += 1

        start = result.smoothedSteps[0]
        end = result.smoothedSteps[1] + 1
        plotData(time=time[start:end],
                 data=data['smoothedMean'],
                 showDataPoint=options.showFittedPoint,
                 color=options.smoothedColor,
                 label='smoothed ' + data['name'])

        if options.showConfidenceInterval:
            upper, lower = getInterval(data['smoothedMean'],
                                       map(abs, data['smoothedVar']),
                                       p=options.confidence)

            plotInterval(time=time[start:end],
                         upper=upper, lower=lower,
                         intervalType=options.intervalType,
                         color=options.smoothedColor)

        plt.legend(loc='best', shadow=True)

# =========================== basic functions ============================


def plotData(time, data, showDataPoint=True, color='black', label='unknown'):
    """
    The function to plot data points.

    Args:
        time: time label
        data: the data points
        showDataPoint: indicate whether scatter plot is needed
        color: the color of the plot
        label: the label for the plot
    """
    if time is None:
        if showDataPoint:
            plt.plot(data, 'o', color=color)
            plt.plot(data, '-', color=color, label=label)
        else:
            plt.plot(data, '-', color=color, label=label)
    else:
        if showDataPoint:
            plt.plot(time, data, 'o', color=color)
            plt.plot(time, data, '-', color=color, label=label)
        else:
            plt.plot(time, data, '-', color=color, label=label)


def plotInterval(time, upper, lower, intervalType, color='black'):
    """
    The function to plot confidence interval.

    Args:
        time: time label
        upper: the upper bound
        lower: the lower bound
        color: the color of the plot
    """
    ALPHA = 0.4
    if time is None:
        if intervalType == 'line':
            plt.plot(upper, '--', color=color)
            plt.plot(lower, '--', color=color)
        elif intervalType == 'ribbon':
            plt.fill_between(upper, lower, facecolor=color,
                             alpha=ALPHA)
    else:
        if intervalType == 'line':
            plt.plot(time, upper, '--', color=color)
            plt.plot(time, lower, '--', color=color)
        elif intervalType == 'ribbon':
            plt.fill_between(time, upper, lower,
                             facecolor=color, alpha=ALPHA)


def plotInitialize():
    """
    Initialize the plot

    """
    plt.figure()


def subplot(size, location):
    """
    Used for plotting multiple figures

    """
    plt.subplot(str(size[0]) + str(size[1]) + str(location))


def plotout():
    """
    Show the plot

    """
    plt.show()


def to1dArray(arrayOf1dMatrix):
    """
    Convert numpy style matrix to usual array

    """
    return [item.tolist()[0][0] for item in arrayOf1dMatrix]



import math

# define the error class for exceptions
class matrixErrors(Exception):
    
    def __init__(self, value):
        self.value = value

    def __str__(self):
        return repr(self.value)


# The class to check matrixErrors
class checker:

    # checking if two matrix has the same dimension
    @staticmethod
    def checkMatrixDimension(A, B):
        if A.shape != B.shape:
            raise matrixErrors('The dimensions do not match!')

    # checking if a vector has the dimension as matrix
    @staticmethod
    def checkVectorDimension(v, A):
        vDim = v.shape
        ADim = A.shape
        if vDim[0] != ADim[0] and vDim[0] != ADim[1] and \
           vDim[1] != ADim[0] and vDim[1] != ADim[1]:
            raise matrixErrors('The dimensions do not match!')

    # checking if a matrix is symmetric
    @staticmethod
    def checkSymmetry(A):
        ADim = A.shape
        if ADim[0] != ADim[1]:
            raise matrixErrors('The matrix is not symmetric!')

        
# a completely unshared copy of lists
def duplicateList(aList):
    if isinstance(aList, list):
        return list(map(duplicateList, aList))
    return aList


# inverse normal cdf function
def rational_approximation(t):
 
    # Abramowitz and Stegun formula 26.2.23.
    # The absolute value of the error should be less than 4.5 e-4.
    c = [2.515517, 0.802853, 0.010328]
    d = [1.432788, 0.189269, 0.001308]
    numerator = (c[2]*t + c[1])*t + c[0]
    denominator = ((d[2]*t + d[1])*t + d[0])*t + 1.0
    return t - numerator / denominator
 
 
def normal_CDF_inverse(p):
    assert p > 0.0 and p < 1
 
    # See article above for explanation of this section.
    if p < 0.5:
        # F^-1(p) = - G^-1(p)
        return -rational_approximation( math.sqrt(-2.0*math.log(p)) )
    else:
        # F^-1(p) = G^-1(1-p)
        return rational_approximation( math.sqrt(-2.0*math.log(1.0-p)) )


def getInterval(means, var, p):
    alpha = abs(normal_CDF_inverse(min(1 - p, p) / 2))
    upper = [0] * len(means)
    lower = [0] * len(means)
    for i in range(0, len(means)):
        upper[i] = means[i] + alpha * math.sqrt(var[i])
        lower[i] = means[i] - alpha * math.sqrt(var[i])

    return (upper, lower)



"""
===========================================================================

The code for autoregressive components

===========================================================================

This code implements the autoregressive component as a sub-class of dynamic.
Different from the dynamic component, the features in the autoReg is generated
from the data, and updated according to the data. All other features are
similar to @dynamic.

"""
from numpy import matrix
from .dynamic import dynamic


class autoReg(dynamic):
    """ The autoReg class alows user to add an autoregressive component to the dlm.
    This code implements the autoregressive component as a sub-class of
    dynamic. Different from the dynamic component, the features in the
    autoReg is generated from the data, and updated according to the data.
    All other features are similar to @dynamic.

    Attributes:
        degree: the degree of autoregressive, i.e., how many days to look back
        data: the time series data used for constructing the autoregressive
              features
        discount factor: the discounting factor
        name: the name of the component
        padding: either 0 or None. The number to be padded for the first degree
                 days, as no previous data is observed to form the feature
                 matrix

    (TODO: change the implementation of autoReg, so that the component uses
     filteredObs as the feature instead of the observed data. To do this, we
     might need pass the dlm as a parameter to autoReg, so that autoReg can
     fetch the filtered observation on the fly)
    """

    def __init__(self,
                 data=None,
                 degree=2,
                 discount=0.99,
                 name='ar2',
                 padding=0):

        if data is None:
            raise NameError('data must be provided to construct' +
                            ' autoregressive component')

        # create fake data to incorporate paddings in the beginning
        fakeData = [padding] * degree + [num for num in data]

        # create features
        features = self.createFeatureMatrix(degree=degree, data=fakeData)

        dynamic.__init__(self,
                         features=features,
                         discount=discount,
                         name=name)
        self.checkDataLength()

        # modify the type to be autoReg
        self.componentType = 'autoReg'

        # the data of the last day has to be stored,
        # since it is not used so far and
        # will be needed when adding new data
        self.lastDay = data[-1]

    def createFeatureMatrix(self, degree, data):
        """ Create the feature matrix based on the supplied data and the degree.

        Args:
            degree: the auto-regressive dependency length.
            data: the raw time series data of the model.
        """

        # we currently don't support missing data for auto regression
        if self.hasMissingData(data):
            raise NameError('The package currently do not support missing ' +
                            'for auto regression. The support has been ' +
                            'implemented, but deprecated due to efficiency ' +
                            'issue. Will support this in next version.')

        # initialize feature matrix
        features = []

        for i in range(degree, len(data)):
            features.append(data[(i - degree):i])

        return features

    # the degree cannot be longer than data
    def checkDataLength(self):
        """ Check whether the degree is less than the time series length

        """
        if self.d >= self.n:
            raise NameError('The degree cannot be longer than the data series')

    # check if there is any none data. We currently don't support missing data
    # for auto regression.
    def hasMissingData(self, aList):
        """ Check whether the list contains None

        """
        for item in aList:
            if item is None:
                return True

        return False

    # overide
    def updateEvaluation(self, date):
        if date < self.n:
            self.evaluation = matrix(self.features[date])
        elif date == self.n:
            self.evaluation = matrix(self.features[-1][1:] + [self.lastDay])
        else:
            raise NameError('The step is out of range')

    # override
    def appendNewData(self, newData):
        """ Append new data to the existing features. Overriding the same method in
        @dynamic

        Args:
            newData: a list of new data

        """
        # fetch the last entry of the feature
        previousDays = self.features[-1]
        fakeData = [num for num in previousDays] + [self.lastDay] + \
                   [num for num in newData]

        # delete the first day which is duplication
        fakeData.pop(0)

        # using the constructed fake data to create new feature sets
        newFeatures = self.createFeatureMatrix(degree=self.d, data=fakeData)

        # append the new feature to the old ones
        self.features.extend(newFeatures)

        # update the last day
        self.lastDay = newData[-1]

        # update n
        self.n = len(self.features)

    # override
    def popout(self, date):
        """ Pop out the data of a specific date and rewrite the correct feature matrix.

        Args:
            date: the index of which to be deleted.

        """

        # if what popped out is the last day, we need to update last day
        if date == self.n - 1:
            self.lastDay = self.features[-1][-1]

            # pop out the corresponding feature
            self.features.pop(self.n - 1)

        # else we can either remove the feature on the given date, but this
        # requires regenerate the new feature matrix, which might be costly.
        # Instead, we directly modify the affected entries in the feature
        # matrix.

        # The change starts from date + 1 to date + degree and
        # we should do it in a reversed order
        else:
            # all the dates that are affected and need to be corrected
            order = list(range(date + 1, min(self.n, date + self.d + 1)))

            # we reverse the processing order
            order.reverse()

            for step in order:
                # popout the deleted date
                self.features[step].pop(self.d - (step - date))

                # padding the data on the last feature to the beginning of
                # today's feature. Note for a 3-autoReg, the feature of day
                # is in a form of [day1, day2, day3]
                self.features[step].insert(0, self.features[step - 1][0])

            # pop out the corresponding feature
            self.features.pop(date)

        # popout the redundent feature and update the length
        self.n -= 1

        # check if the degree is longer than the data series
        self.checkDataLength()

    def alter(self, date, dataPoint):
        """ Alter the data of a particular date, and change the corresponding
            feature matrix.

        Args:
           date: The date to be modified.
           dataPoint: The new dataPoint to be filled in.

        """
        # if what modified is the last day, we need to update last day
        if date == self.n - 1:
            self.lastDay = dataPoint

        # else we can either modify the feature on the given date, but this
        # requires regenerate the new feature matrix, which might be costly.
        # Instead, we directly modify the affected entries in the feature
        # matrix.

        # The change starts from date + 1 to date + degree and
        # we should do it in a reversed order
        else:
            # all the dates that are affected and need to be corrected
            order = list(range(date + 1, min(self.n, date + self.d + 1)))

            # we reverse the processing order
            order.reverse()

            for step in order:
                # popout the deleted date
                self.features[step][self.d - (step - date)] = dataPoint

        # check if the degree is longer than the data series
        self.checkDataLength()



# This is the PyDLM package

__all__ = ['dlm', 'trend', 'seasonality', 'dynamic', 'autoReg', 'longSeason']

from pydlm.dlm import dlm
from pydlm.modeler.trends import trend
from pydlm.modeler.seasonality import seasonality
from pydlm.modeler.dynamic import dynamic
from pydlm.modeler.autoReg import autoReg
from pydlm.modeler.longSeason import longSeason



"""
===========================================================================

The code for long seasonality components

===========================================================================

This code implements the long seasonality component as a sub-class of dynamic.
The difference between the long seasonality is that
1) The seasonality component use each date as a unit and change in a given
periodicy. For example, 1, 2, 3, 4, 1, 2, 3, 4.
2) However, the long seasonality is capable to group a couple of days as the
basic unit and change in a periodicy. For example, 1, 1, 1, 2, 2, 2, 3, 3, 3,
4, 4, 4.
The usecase for long seasonality is to model longer seasonality with the
short-term seasonality. For example, the short-term seansonality can be
used to model the day of a weak patten and the long seasonality can be used
to model the week of a month patten in the same model.
Different from the dynamic component, the features in the autoReg is generated
from the data, and updated according to the data. All other features are
similar to @dynamic.

"""
from numpy import matrix
from .dynamic import dynamic


class longSeason(dynamic):
    """ The longSeason class alows user to add a long seasonality component
    to the dlm. The difference between the long seasonality is that
    1) The seasonality component use each date as a unit and change in a given
    periodicity. For example, 1, 2, 3, 4, 1, 2, 3, 4.
    2) However, the long seasonality is capable to group couple of days as the
    basic unit and change in a periodicity. For example, 1, 1, 1, 2, 2, 2,
    3, 3, 3, 4, 4, 4.
    The usecase for long seasonality is to model longer seasonality with the
    short-term seasonality. For example, the short-term seansonality can be
    used to model the day of a weak patten and the long seasonality can be used
    to model the week of a month patten in the same model.
    This code implements the longSeason component as a sub-class of
    dynamic. Different from the dynamic component, the features in the
    autoReg is generated from the data, and updated according to the data.
    All other features are similar to @dynamic.

    Attributes:
        period: the periodicity, i.e., how many different states it has in
                one period
        stay: the length of a state last.
        discount factor: the discounting factor
        name: the name of the component

    """

    def __init__(self,
                 data=None,
                 period=4,
                 stay=7,
                 discount=0.99,
                 name='longSeason'):

        self.period = period
        self.stay = stay
        if data is None:
            raise NameError('Data must be provided for longSeason.')

        # create features. nextState and state are used to
        # remember the next feature shap
        features, self.nextState = self.createFeatureMatrix(period=period,
                                                            stay=stay,
                                                            n=len(data),
                                                            state=[0, 0])

        dynamic.__init__(self,
                         features=features,
                         discount=discount,
                         name=name)
        self.checkDataLength()

        # modify the type to be autoReg
        self.componentType = 'longSeason'

    def createFeatureMatrix(self, period, stay, n, state):
        """ Create the feature matrix based on the supplied data and the degree.

        Args:
            period: the periodicity of the component
            stay: the length of the base unit, i.e, how long before change to
                  change to the next state.
        """

        # initialize feature matrix
        nextState = state
        features = []

        count = 0
        while count < n:
            # create new feature for next state
            new_feature = [0] * period
            new_feature[nextState[0]] = 1
            features.append(new_feature)

            # update the state
            nextState[1] = (nextState[1] + 1) % stay
            if nextState[1] == 0:
                nextState[0] = (nextState[0] + 1) % period

            count += 1

        return features, nextState

    # the degree cannot be longer than data
    def checkDataLength(self):
        """ Check whether the degree is less than the time series length

        """
        if self.d >= self.n:
            raise NameError('The degree cannot be longer than the data series')

    # override
    def appendNewData(self, newData):
        """ Append new data to the existing features. Overriding the same method in
        @dynamic

        Args:
            newData: a list of new data

        """
        # create the new features
        incrementLength = len(newData) + self.n - len(self.features)
        if incrementLength > 0:
            newFeatures, \
                self.nextState = self.createFeatureMatrix(period=self.period,
                                                          stay=self.stay,
                                                          n=incrementLength,
                                                          state=self.nextState)
        self.features.extend(newFeatures)
        self.n += len(newData)

    # override
    def popout(self, date):
        """ Pop out the data of a specific date and rewrite the correct feature matrix.

        Args:
            date: the index of which to be deleted.

        """
        # Since the seasonality is a fixed patten,
        # no matter what date is popped out
        # we just need to remove the last date,
        # otherwise the feature patten will be
        # changed.

        # if you want to delete a date and change
        # the underlying patten, i.e., shorten
        # the periodicity of the period that date
        # is presented, you should use ignore
        # instead
        print('Popout the date will change the whole' +
              ' seasonality patten on all the' +
              'future days. If you want to keep the' +
              ' seasonality patten on the future' +
              'days unchanged. Please use ignore instead')

        self.features.pop()
        self.n -= 1

        # push currentState back by 1 day. Need to take care of all
        # corner cases.
        if self.nextState[1] == 0:
            self.nextState[1] = self.stay - 1
            if self.nextState[0] == 0:
                self.nextState[0] = self.period - 1
            else:
                self.nextState[0] -= 1
        else:
            self.nextState[1] -= 1

    def alter(self, date, dataPoint):
        """ We do nothing to longSeason, when altering the main data

        """

        # do nothing
        pass

    def updateEvaluation(self, step):
        """ update the evaluation matrix to a specific date
        This function is used when fitting the forward filter and
        backward smoother
        in need of updating the correct evaluation matrix

        """
        if step < len(self.features):
            self.evaluation = matrix(self.features[step])
        else:
            newFeatures, \
                self.nextState = self.createFeatureMatrix(
                    period=self.period,
                    stay=self.stay,
                    n=step + 1 - len(self.features),
                    state=self.nextState)
            self.features.extend(newFeatures)
            self.evaluation = matrix(self.features[step])
        self.step = step



# -*- coding: utf-8 -*-
"""
===============================================================================

The code for the class dlm

===============================================================================

This is the main class of the Bayeisan dynamic linear model.
It provides the modeling, filtering, forecasting and smoothing function
 of a dlm. The dlm use the @builder to construct the @baseModel based
 on user supplied @components and then run @kalmanFilter to filter the result.

Example:
>>> # randomly generate fake data on 1000 days
>>> import numpy as np
>>> data = np.random.random((1, 1000))

>>> # construct the dlm of a linear trend and a 7-day seasonality
>>> from pydlm import dlm, trend, seasonality
>>> myDlm = dlm(data) + trend(degree = 2, 0.98) + seasonality(period = 7, 0.98)

>>> # filter the result
>>> myDlm.fitForwardFilter()

>>> # extract the filtered result
>>> myDlm.getFilteredObs()

"""
# This is the major class for fitting time series data using the
# dynamic linear model. dlm is a subclass of builder, with adding the
# Kalman filter functionality for filtering the data

from copy import deepcopy
from pydlm.func._dlm import _dlm
from pydlm.base.tools import getInterval


class dlm(_dlm):
    """ The main class of the dynamic linear model.


    This is the main class of the Bayeisan dynamic linear model.
    It provides the modeling, filtering, forecasting and smoothing
    function of a dlm.
    The dlm use the @builder to construct the @baseModel based on user supplied
    @components and then run @kalmanFilter to filter the result.

    Example 1:
        >>> # randomly generate fake data on 1000 days
        >>> import numpy as np
        >>> data = np.random.random((1, 1000))
        >>> # construct the dlm of a linear trend and a 7-day seasonality
        >>> myDlm = dlm(data) + trend(degree = 2, 0.98) + seasonality(period = 7, 0.98)
        >>> # filter the result
        >>> myDlm.fitForwardFilter()
        >>> # extract the filtered result
        >>> myDlm.getFilteredObs()

    Example 2 (fit a linear regression):
        >>> from pydlm import dynamic
        >>> data = np.random.random((1, 100))
        >>> mydlm = dlm(data) + trend(degree=1, 0.98, name='a') +
                        dynamic(features=[[i] for i in range(100)], 1, name='b')
        >>> mydlm.fit()
        >>> coef_a = mydlm.getLatentState('a')
        >>> coef_b = mydlm.getLatentState('b')

    Attributes:
       data: a list of doubles of the raw time series data.
             It could be either the python's built-in list of
             doubles or numpy 1d array.

    """
    # define the basic members
    # initialize the result
    def __init__(self, data):
        _dlm.__init__(self, data)

        # indicate whether the plot modules has been loaded.
        # We add this flag, since we only import plot module
        # when they are called to avoid any error due to
        # plot that blocks using this package. This flag can
        # help without doing import-check (expensive) every time
        # when plot function is called.
        self.plotLibLoaded = False
# ===================== modeling components =====================

    # add component
    def add(self, component):
        """ Add new modeling component to the dlm.

        Currently support: trend, seasonality, autoregression
        and dynamic component.

        Args:
            component: the modeling component, could be either one
                       of the following:\n
                       trend, seasonality, dynamic, autoReg.

        Returns:
            A dlm object with added component.

        """
        self.__add__(component)

    def __add__(self, component):
        self.builder.__add__(component)
        self.initialized = False
        return self

    # list all components
    def ls(self):
        """ List out all existing components

        """
        self.builder.ls()

    # delete one component
    def delete(self, name):
        """ Delete model component by its name

        Args:
            name: the name of the component.

        """
        self.builder.delete(name)
        self.initialized = False

# ========================== model training component =======================

    def fitForwardFilter(self, useRollingWindow=False, windowLength=3):
        """ Fit forward filter on the available data.

        User can choose use rolling windowFront
        or not. If user choose not to use the rolling window,
        then the filtering will be based on all the previous data.
        If rolling window is used, then the filtering for a particular
        date will only consider previous dates that are
        within the rolling window length.

        Args:
            useRollingWindow: indicate whether rolling window should be used.
            windowLength: the length of the rolling window if used.

        """
        # check if the feature size matches the data size
        self._checkFeatureSize()

        # see if the model has been initialized
        if not self.initialized:
            self._initialize()

        if self._printInfo:
            print('Starting forward filtering...')
        if not useRollingWindow:
            # we start from the last step of previous fitering
            if self.result.filteredType == 'non-rolling':
                start = self.result.filteredSteps[1] + 1
            else:
                start = 0
                # because we refit the forward filter, we need to reset the
                # backward smoother as well.
                self.result.smoothedSteps = [0, -1]

            # determine whether renew should be used
            self._forwardFilter(start=start,
                                end=self.n - 1,
                                renew=self.options.stable)
            self.result.filteredType = 'non-rolling'
        else:
            if self.result.filteredType == 'rolling':
                windowFront = self.result.filteredSteps[1] + 1
            else:
                windowFront = 0
                # because we refit the forward filter, we need to reset the
                # backward smoother as well.
                self.result.smoothedSteps = [0, -1]

            self.result.filteredType = 'rolling'
            # if end is still within (0, windowLength - 1), we should run the
            # usual ff from
            if windowFront < windowLength:
                self._forwardFilter(start=self.result.filteredSteps[1] + 1,
                                    end=min(windowLength - 1, self.n - 1))

            # for the remaining date, we use a rolling window
            for today in range(max(windowFront, windowLength), self.n):
                self._forwardFilter(start=today - windowLength + 1,
                                    end=today,
                                    save=today,
                                    ForgetPrevious=True)

        self.result.filteredSteps = [0, self.n - 1]
        self.turnOn('filtered plot')
        self.turnOn('predict plot')

        # reset everything that needs reset
        self._clean()

        if self._printInfo:
            print('Forward fitering completed.')

    def fitBackwardSmoother(self, backLength=None):
        """ Fit backward smoothing on the data. Starting from the last observed date.

        Args:
            backLength: integer, indicating how many days the backward smoother
            should go, starting from the last date.

        """

        # see if the model has been initialized
        if not self.initialized:
            raise NameError('Backward Smoother has to be run after' +
                            ' forward filter')

        if self.result.filteredSteps[1] != self.n - 1:
            raise NameError('Forward Fiter needs to run on full data before' +
                            'using backward Smoother')

        # default value for backLength
        if backLength is None:
            backLength = self.n

        if self._printInfo:
            print('Starting backward smoothing...')
        # if the smoothed dates has already been done, we do nothing
        if self.result.smoothedSteps[1] == self.n - 1 and \
           self.result.smoothedSteps[0] <= self.n - 1 - backLength + 1:
            return None

        # if the smoothed dates start from n - 1, we just need to continue
        elif self.result.smoothedSteps[1] == self.n - 1:
            self._backwardSmoother(start=self.result.smoothedSteps[0] - 1,
                                   days=backLength)

        # if the smoothed dates are even earlier,
        # we need to start from the beginning
        elif self.result.smoothedSteps[1] < self.n - 1:
            self._backwardSmoother(start=self.n - 1, days=backLength)

        self.result.smoothedSteps = [self.n - backLength, self.n - 1]
        self.turnOn('smoothed plot')

        # reset everything that needs reset
        self._clean()

        if self._printInfo:
            print('Backward smoothing completed.')

    def fit(self):
        """ An easy caller for fitting both the forward filter and backward smoother.

        """
        self.fitForwardFilter()
        self.fitBackwardSmoother()

# =========================== model prediction ==============================

    # One day ahead prediction function
    def predict(self, date=None, featureDict=None):
        """ One day ahead predict based on the current data for a specific future
        days.

        The predict result is based on all the data before date and predict the
        observation at date + days.

        Args:
            date: the index when the prediction based on. Default to the
                  last day.
            featureDict: the feature set for the dynamic Components, in a form
                  of {"component_name": feature}. If the featureDict is not
                  supplied, then the algo reuse those stored in the dynamic
                  components. For dates beyond the last day, featureDict must
                  be supplied.

        Returns:
            A tuple. (Predicted observation, variance of the predicted
            observation)

        """
        # the default prediction date
        if date is None:
            date = self.n - 1

        # check if the data on the date has been filtered
        if date > self.result.filteredSteps[1]:
            raise NameError('Prediction can only be made right' +
                            ' after the filtered date')

        return self._oneDayAheadPredict(date=date, featureDict=featureDict)

    def continuePredict(self, featureDict=None):
        """ Continue prediction after the one-day ahead predict.

        If users want to have a multiple day prediction, they can opt to use
        continuePredict after predict with new features contained in
        featureDict. For example,

        >>> # predict 3 days after the last day
        >>> myDLM.predict(featureDict=featureDict_day1)
        >>> myDLM.continuePredict(featureDict=featureDict_day2)
        >>> myDLM.continuePredict(featureDict=featureDict_day3)

        Args:
            featureDict: the feature set for the dynamic components, stored
                         in a for of {"component name": feature}. If the set
                         was not supplied, then the algo will re-use the old
                         feature. For days beyond the data, the featureDict
                         for every dynamic component must be provided.

        Returns:
            A tupe. (predicted observation, variance)
        """
        if self.result.predictStatus is None:
            raise NameError('continuePredict has to come after predict.')

        return self._continuePredict(featureDict=featureDict)

# =========================== result components =============================

    def getAll(self):
        """ get all the _result class which contains all results

        Returns:
            The @result object containing all computed results.

        """
        return deepcopy(self.result)

    def getMean(self, filterType='forwardFilter', name='main'):
        """ get mean for data or component.

        If the working dates are not (0, self.n - 1),
        then a warning will prompt stating the actual filtered dates.

        Args:
            filterType: the type of mean to be returned. Could be
                        'forwardFilter', 'backwardSmoother', and 'predict'.
                        Default to 'forwardFilter'.
            name: the component to get mean. When name = 'main', then it
                  returns the filtered mean for the time series. When
                  name = some component's name, then it returns the filtered
                  mean for that component. Default to 'main'.

        Returns:
            A list of the time series observations based on the choice

        """
        # get the working date
        start, end = self._checkAndGetWorkingDates(filterType=filterType)

        # get the mean for the fitlered data
        if name == 'main':
            # get out of the matrix form
            if filterType == 'forwardFilter':
                return self._1DmatrixToArray(
                    self.result.filteredObs[start:end])
            elif filterType == 'backwardSmoother':
                return self._1DmatrixToArray(
                    self.result.smoothedObs[start:end])
            elif filterType == 'predict':
                return self._1DmatrixToArray(
                    self.result.predictedObs[start:end])
            else:
                raise NameError('Incorrect filter type.')

        # get the mean for the component
        self._checkComponent(name)
        return self._getComponentMean(name=name,
                                      filterType=filterType,
                                      start=start, end=end)

    def getVar(self, filterType='forwardFilter', name='main'):
        """ get the variance for data or component.

        If the filtered dates are not (0, self.n - 1),
        then a warning will prompt stating
        the actual filtered dates.

        Args:
            filterType: the type of variance to be returned. Could be
                        'forwardFilter', 'backwardSmoother', and 'predict'.
                        Default to 'forwardFilter'.
            name: the component to get variance. When name = 'main', then it
                  returns the filtered variance for the time series. When
                  name = some component's name, then it returns the filtered
                  variance for that component. Default to 'main'.

        Returns:
            A list of the filtered variances based on the choice.

        """
        # get the working date
        start, end = self._checkAndGetWorkingDates(filterType=filterType)

        # get the variance for the time series data
        if name == 'main':
            # get out of the matrix form
            if filterType == 'forwardFilter':
                return self._1DmatrixToArray(
                    self.result.filteredObsVar[start:end])
            elif filterType == 'backwardSmoother':
                return self._1DmatrixToArray(
                    self.result.smoothedObsVar[start:end])
            elif filterType == 'predict':
                return self._1DmatrixToArray(
                    self.result.predictedObsVar[start:end])
            else:
                raise NameError('Incorrect filter type.')

        # get the variance for the component
        self._checkComponent(name)
        return self._getComponentVar(name=name, filterType=filterType,
                                     start=start, end=end)

    def getInterval(self, p=0.95, filterType='forwardFilter', name='main'):
        """ get the confidence interval for data or component.

        If the filtered dates are not
        (0, self.n - 1), then a warning will prompt stating the actual
        filtered dates.

        Args:
            p: The confidence level.
            filterType: the type of CI to be returned. Could be
                        'forwardFilter', 'backwardSmoother', and 'predict'.
                        Default to 'forwardFilter'.
            name: the component to get CI. When name = 'main', then it
                  returns the confidence interval for the time series. When
                  name = some component's name, then it returns the confidence
                  interval for that component. Default to 'main'.

        Returns:
            A tuple with the first element being a list of upper bounds
            and the second being a list of the lower bounds.

        """
        # get the working date
        start, end = self._checkAndGetWorkingDates(filterType=filterType)

        # get the mean and the variance for the time series data
        if name == 'main':
            # get out of the matrix form
            if filterType == 'forwardFilter':
                compMean = self._1DmatrixToArray(
                    self.result.filteredObs[start:end])
                compVar = self._1DmatrixToArray(
                    self.result.filteredObsVar[start:end])
            elif filterType == 'backwardSmoother':
                compMean = self._1DmatrixToArray(
                    self.result.smoothedObs[start:end])
                compVar = self._1DmatrixToArray(
                    self.result.smoothedObsVar[start:end])
            elif filterType == 'predict':
                compMean = self._1DmatrixToArray(
                    self.result.predictedObs[start:end])
                compVar = self._1DmatrixToArray(
                    self.result.predictedObsVar[start:end])
            else:
                raise NameError('Incorrect filter type.')

        # get the mean and variance for the component
        else:
            self._checkComponent(name)
            compMean = self._getComponentMean(name=name,
                                              filterType=filterType,
                                              start=start, end=end)
            compVar = self._getComponentVar(name=name,
                                            filterType=filterType,
                                            start=start, end=end)

        # get the upper and lower bound
        upper, lower = getInterval(compMean, compVar, p)
        return (upper, lower)

    def getLatentState(self, filterType='forwardFilter', name='all'):
        """ get the latent states for different components and filters.

        If the filtered dates are not (0, self.n - 1),
        then a warning will prompt stating the actual filtered dates.

        Args:
            filterType: the type of latent states to be returned. Could be
                        'forwardFilter', 'backwardSmoother', and 'predict'.
                        Default to 'forwardFilter'.
            name: the component to get latent state. When name = 'all', then it
                  returns the latent states for the time series. When
                  name = some component's name, then it returns the latent
                  states for that component. Default to 'all'.

        Returns:
            A list of lists, standing for the latent states given
            the different choices.

        """
        # get the working dates
        start, end = self._checkAndGetWorkingDates(filterType=filterType)

        # to return the full latent states
        if name == 'all':
            if filterType == 'forwardFilter':
                return map(lambda x: x if x is None
                           else self._1DmatrixToArray(x),
                           self.result.filteredState[start:end])
            elif filterType == 'backwardSmoother':
                return map(lambda x: x if x is None
                           else self._1DmatrixToArray(x),
                           self.result.smoothedState[start:end])
            elif filterType == 'predict':
                return map(lambda x: x if x is None
                           else self._1DmatrixToArray(x),
                           self.result.smoothedState[start:end])
            else:
                raise NameError('Incorrect filter type.')

        # to return the latent state for a given component
        self._checkComponent(name)
        return map(lambda x: x if x is None else self._1DmatrixToArray(x),
                   self._getLatentState(name=name, filterType=filterType,
                                        start=start, end=end))

    def getLatentCov(self, filterType='forwardFilter', name='all'):
        """ get the error covariance for different components and
        filters.

        If the filtered dates are not (0, self.n - 1),
        then a warning will prompt stating the actual filtered dates.

        Args:
            filterType: the type of latent covariance to be returned. Could be
                        'forwardFilter', 'backwardSmoother', and 'predict'.
                        Default to 'forwardFilter'.
            name: the component to get latent cov. When name = 'all', then it
                  returns the latent covariance for the time series. When
                  name = some component's name, then it returns the latent
                  covariance for that component. Default to 'all'.

        Returns:
            A list of numpy matrices, standing for the filtered latent
            covariance.

        """
        # get the working dates
        start, end = self._checkAndGetWorkingDates(filterType=filterType)

        # to return the full latent covariance
        if name == 'all':
            if filterType == 'forwardFilter':
                return self.result.filteredCov[start:end]
            elif filterType == 'backwardSmoother':
                return self.result.smoothedCov[start:end]
            elif filterType == 'predict':
                return self.result.smoothedCov[start:end]
            else:
                raise NameError('Incorrect filter type.')

        # to return the latent covariance for a given component
        self._checkComponent(name)
        return self._getLatentCov(name=name, filterType=filterType,
                                  start=start, end=end)

# ======================= data appending, popping and altering ===============

    # Append new data or features to the dlm
    def append(self, data, component='main'):
        """ Append the new data to the main data or the components (new feature data)

        Args:
            data: the new data
            component: the name of which the new data to be added to.\n
                       'main': the main time series data\n
                       other omponent name: add new feature data to other
                       component.

        """
        # if we are adding new data to the time series
        if component == 'main':
            # add the data to the self.data
            self.data.extend(list(data))

            # update the length
            self.n += len(data)
            self.result._appendResult(len(data))

            # update the automatic components as well
            for component in self.builder.automaticComponents:
                comp = self.builder.automaticComponents[component]
                comp.appendNewData(data)

            # give a warning to remind to append dynamic components
            if len(self.builder.dynamicComponents) > 0:
                print('Remember to append the new features for the' +
                      ' dynamic components as well')

        # if we are adding new data to the features of dynamic components
        elif component in self.builder.dynamicComponents:
            comp = self.builder.dynamicComponents[component]
            comp.appendNewData(data)

        else:
            raise NameError('Such dynamic component does not exist.')

        # reset everything that needs reset
        self._clean()

    # pop the data of a specific date out
    def popout(self, date):
        """ Pop out the data for a given date

        Args:
            date: the index indicates which date to be popped out.

        """
        if date < 0 or date > self.n - 1:
            raise NameError('The date should be between 0 and ' +
                            str(self.n - 1))

        # pop out the data at date
        self.data.pop(date)
        self.n -= 1

        # pop out the feature at date
        for name in self.builder.dynamicComponents:
            comp = self.builder.dynamicComponents[name]
            comp.popout(date)

        # pop out the results at date
        self.result._popout(date)

        # update the filtered and the smoothed steps
        self.result.filteredSteps[1] = date - 1
        self.result.smoothedSteps[1] = date - 1

        if self.result.filteredSteps[0] > self.result.filteredSteps[1]:
            self.result.filteredSteps = [0, -1]
            self.result.smoothedSteps = [0, -1]

        elif self.result.smoothedSteps[0] > self.result.smoothedSteps[1]:
            self.result.smoothedSteps = [0, -1]

        # reset everything that needs reset
        self._clean()

    # alter the data of a specific days
    def alter(self, date, data, component='main'):
        """ To alter the data for a specific date and a specific component.

        Args:
            date: the date of the altering data
            data: the new data
            component: the component for which the new data need to be
                       supplied to.\n
                       'main': the main time series data\n
                       other component name: other component feature data

        """
        if date < 0 or date > self.n - 1:
            raise NameError('The date should be between 0 and ' +
                            str(self.n - 1))

        # to alter the data for the observed chain
        if component == 'main':
            self.data[date] = data

            # we also automatically alter all the automatic components
            for component in self.builder.automaticComponents:
                comp = self.builder.automaticComponents[component]
                comp.alter(date, data)

        # to alter the feature of a component
        elif component in self.builder.dynamicComponents:
            comp = self.builder.dynamicComponents[component]
            comp.features[component] = data

        else:
            raise NameError('Such dynamic component does not exist.')

        # update the filtered and the smoothed steps
        self.result.filteredSteps[1] = date - 1
        self.result.smoothedSteps[1] = date - 1

        if self.result.filteredSteps[0] > self.result.filteredSteps[1]:
            self.result.filteredSteps = [0, -1]
            self.result.smoothedSteps = [0, -1]

        elif self.result.smoothedSteps[0] > self.result.smoothedSteps[1]:
            self.result.smoothedSteps = [0, -1]

        # reset everything that needs reset
        self._clean()

    # ignore the data of a given date
    def ignore(self, date):
        """ Ignore the data for a specific day. treat it as missing data

        Args:
            date: the date to ignore.

        """
        if date < 0 or date > self.n - 1:
            raise NameError('The date should be between 0 and ' +
                            str(self.n - 1))

        self.alter(date=date, data=None, component='main')

# ============================= plot component =========================

    def turnOn(self, switch):
        """ "turn on" Operation for the dlm plotting options.

        Args:
            switch: The key word to switch on. \n
                    'filtered plot', 'filter' to plot filtered results\n
                    'smoothed plot', 'smooth' to plot smoothed results\n
                    'predict plot', 'predict', to plot one-step ahead results\n
                    'confidence interval', 'confierence', 'CI' to plot CI's\n
                    'data points', 'data', 'data point' to plot original data\n
                    'multiple', 'separate' to plot results in separate
                    figures\n
                    'fitted dots', 'fitted' to plot fitted results with dots
        """
        if switch in set(['filtered plot', 'filter',
                          'filtered results', 'filtering']):
            self.options.plotFilteredData = True
        elif switch in set(['smoothed plot', 'smooth',
                            'smoothed results', 'smoothing']):
            self.options.plotSmoothedData = True
        elif switch in set(['predict plot', 'predict',
                            'predicted results', 'prediction']):
            self.options.plotPredictedData = True
        elif switch in set(['confidence interval', 'confidence',
                            'interval', 'CI', 'ci']):
            self.options.showConfidenceInterval = True
        elif switch in set(['data points', 'data point', 'points', 'data']):
            self.options.showDataPoint = True
        elif switch in set(['multiple', 'multiple plots',
                            'separate plots', 'separate']):
            self.options.separatePlot = True
        elif switch in set(['fitted dots', 'fitted results',
                            'fitted data', 'fitted']):
            self.options.showFittedPoint = True
        else:
            raise NameError('no such options')

    def turnOff(self, switch):
        """ "turn off" Operation for the dlm plotting options.

        Args:
            switch: The key word to switch off. \n
                    'filtered plot', 'filter' to not plot filtered results\n
                    'smoothed plot', 'smooth' to not plot smoothed results\n
                    'predict plot', 'predict', to not plot one-step ahead
                    results\n
                    'confidence interval', 'confierence', 'CI'
                    to not plot CI's\n
                    'data points', 'data', 'data point' to not
                    plot original data\n
                    'multiple', 'separate' to not plot results
                    in separate figures\n
                    'fitted dots', 'fitted' to not plot fitted
                    results with dots

        """
        if switch in set(['filtered plot', 'filter', 'filtered results',
                          'filtering']):
            self.options.plotFilteredData = False
        elif switch in set(['smoothed plot', 'smooth', 'smoothed results',
                            'smoothing']):
            self.options.plotSmoothedData = False
        elif switch in set(['predict plot', 'predict', 'predicted results',
                            'prediction']):
            self.options.plotPredictedData = False
        elif switch in set(['confidence interval', 'confidence', 'interval',
                            'CI', 'ci']):
            self.options.showConfidenceInterval = False
        elif switch in set(['data points', 'data point', 'points', 'data']):
            self.options.showDataPoint = False
        elif switch in set(['multiple', 'multiple plots', 'separate plots',
                            'separate']):
            self.options.separatePlot = False
        elif switch in set(['fitted dots', 'fitted results', 'fitted data',
                            'fitted']):
            self.options.showFittedPoint = False
        else:
            raise NameError('no such options')

    def setColor(self, switch, color):
        """ "set" Operation for the dlm plotting colors

        Args:
            switch: key word. Controls over
                    filtered/smoothed/predicted results,
            color: the color for the corresponding keyword.
        """
        if switch in set(['filtered plot', 'filter', 'filtered results',
                          'filtering']):
            self.options.filteredColor = color
        elif switch in set(['smoothed plot', 'smooth', 'smoothed results',
                            'smoothing']):
            self.options.smoothedColor = color
        elif switch in set(['predict plot', 'predict', 'predicted results',
                            'prediction']):
            self.options.predictedColor = color
        elif switch in set(['data points', 'data point', 'points', 'data']):
            self.options.dataColor = color
        else:
            raise NameError('no such options')

    def setConfidence(self, p=0.95):
        """ Set the confidence interval for the plot

        """
        assert p >= 0 and p <= 1
        self.options.confidence = p

    def setIntervalType(self, intervalType):
        """ Set the confidence interval type

        """
        if intervalType == 'ribbon' or intervalType == 'line':
            self.options.intervalType = intervalType
        else:
            raise NameError('No such type for confidence interval.')

    def resetPlotOptions(self):
        """ Reset the plotting option for the dlm class

        """
        self.options.plotOriginalData = True
        self.options.plotFilteredData = True
        self.options.plotSmoothedData = True
        self.options.plotPredictedData = True
        self.options.showDataPoint = True
        self.options.showFittedPoint = False
        self.options.showConfidenceInterval = True
        self.options.dataColor = 'black'
        self.options.filteredColor = 'blue'
        self.options.predictedColor = 'green'
        self.options.smoothedColor = 'red'
        self.options.separatePlot = True
        self.options.confidence = 0.95
        self.options.intervalType = 'ribbon'

    # plot the result according to the options
    def plot(self, name='main'):
        """ The main plot function. The dlmPlot and the matplotlib will only be loaded
        when necessary.

        Args:
            name: component to plot. Default to 'main', in which we plot the
                  filtered time series. If a component name is given
                  It plots the mean of the component, i.e., the observed value
                  that attributes to that particular component, which equals to
                  evaluation * latent states for that particular component.

        """

        # load the library only when needed
        # import pydlm.plot.dlmPlot as dlmPlot
        self.loadPlotLibrary()

        if self.time is None:
            time = range(len(self.data))

        # initialize the figure
        dlmPlot.plotInitialize()

        # change option setting if some results are not available
        if not self.initialized:
            raise NameError('The model must be constructed and' +
                            ' fitted before ploting.')

        # check the filter status and automatically turn off bad plots
        self._checkPlotOptions()

        # plot the main time series after filtering
        if name == 'main':
            # if we just need one plot
            if self.options.separatePlot is not True:
                dlmPlot.plotInOneFigure(time=time,
                                        data=self.data,
                                        result=self.result,
                                        options=self.options)
            # otherwise, we plot in multiple figures
            else:
                dlmPlot.plotInMultipleFigure(time=time,
                                             data=self.data,
                                             result=self.result,
                                             options=self.options)

        # plot the component after filtering
        elif self._checkComponent(name):
            # create the data for ploting
            data = {}
            if self.options.plotFilteredData:
                data['filteredMean'] = self.getMean(
                    filterType='forwardFilter', name=name)
                data['filteredVar'] = self.getVar(
                    filterType='forwardFilter', name=name)

            if self.options.plotSmoothedData:
                data['smoothedMean'] = self.getMean(
                    filterType='backwardSmoother', name=name)
                data['smoothedVar'] = self.getVar(
                    filterType='backwardSmoother', name=name)

            if self.options.plotPredictedData:
                data['predictedMean'] = self.getMean(
                    filterType='predict', name=name)
                data['predictedVar'] = self.getVar(
                    filterType='predict', name=name)

            if len(data) == 0:
                raise NameError('Nothing is going to be drawn, due to ' +
                                'user choices.')
            data['name'] = name
            dlmPlot.plotComponent(time=time,
                                  data=data,
                                  result=self.result,
                                  options=self.options)
        dlmPlot.plotout()

    def plotCoef(self, name, dimensions=None):
        """ Plot function for the latent states (coefficents of dynamic
        component).

        Args:
            name: the name of the component to plot.
                  It plots the latent states for the component. If dimension of
                  the given component is too high, we truncate
                  to the first five. Or the user can supply the ideal
                  dimensions for plot in the dimensions parameter.
            dimensions: dimensions will be used
                        as the indexes to plot within that component latent
                        states.
        """
        # load the library only when needed
        # import pydlm.plot.dlmPlot as dlmPlot
        self.loadPlotLibrary()

        # provide a fake time for plotting
        if self.time is None:
            time = range(len(self.data))

        # change option setting if some results are not available
        if not self.initialized:
            raise NameError('The model must be constructed and' +
                            ' fitted before ploting.')

        # check the filter status and automatically turn off bad plots
        self._checkPlotOptions()

        # plot the latent states for a given component
        if self._checkComponent(name):

            # find its coordinates in the latent state
            indx = self.builder.componentIndex[name]
            # get the real latent states
            coordinates = range(indx[0], (indx[1] + 1))
            # if user supplies the dimensions
            if dimensions is not None:
                coordinates = [coordinates[i] for i in dimensions]

            # otherwise, if there are too many latent states, we
            # truncated to the first five.
            elif len(coordinates) > 5:
                coordinates = coordinates[:5]

            dlmPlot.plotLatentState(time=time,
                                    coordinates=coordinates,
                                    result=self.result,
                                    options=self.options,
                                    name=name)
        else:
            raise NameError('No such component.')

        dlmPlot.plotout()
# ================================ control options =========================

    def showOptions(self):
        """ Print out all the option values

        """
        allItems = vars(self.options)
        for item in allItems:
            print(item + ': ' + str(allItems[item]))

    def stableMode(self, use=True):
        """ Turn on the stable mode, i.e., using the renewal strategy.

            Indicate whether the renew strategy should be used to add numerical
            stability. When the filter goes over certain steps,
            the information contribution of the previous data has decayed
            to minimum. In the stable mode, We then ignore those days and
            refit the time series starting from current - renewTerm, where
            renewTerm is computed according to the discount. Thus,
            the effective sample size of the dlm is twice
            renewTerm. When discount = 1, there will be no renewTerm,
            since all the information will be passed along.
        """
        if self.options.stable != use:
            self.initialized = False
        self.options.stable = use

    def loadPlotLibrary(self):
        if not self.plotLibLoaded:
            global dlmPlot
            import pydlm.plot.dlmPlot as dlmPlot
            self.plotLibLoaded = True



"""
===============================================================================

The code for all hidden method for the class dlm

===============================================================================

This piece of code include all the hidden methods and members of the class dlm.
It provides the basic modeling, filtering, forecasting and smoothing of a dlm.

"""
from numpy import matrix
from numpy import dot
from pydlm.base.kalmanFilter import kalmanFilter
from pydlm.modeler.builder import builder

# this class defines the basic functionalities for dlm, which is not supposed
# to be used by the user. Most functionality in the main dlm will be
# constructed by using the hidden functions in this class


class _dlm:
    """ _dlm includes all hidden functions that used by the class dlm. These hidden
    methods provide the basic modeling, filtering, forecasting and smoothing of
    dlm.

    Attributes:
        data: the observed time series data
        n: the length of the time series data
        result: the inner class that records the filtered and smoothed results
        builder: the @builder that is used for providing the modeling
        functionality. For details please refer to @builder
        Filter: the filter used for filtering time series data using the model.
                For details please refer to @kalmanFilter
        initialized: indicates whether the dynamic linear model has been
                     initialized
        options: model options, including initial guess of the observational
                 variance.
                 More is going to be added (plot options and shrinkage options)
        time: the time label, used for plotting


    Methods:
        _initialize: initialize the dlm (builder and kalmanFilter)
        _forwardFilter: run forward filter for a specific start and end date
        _backwardSmoother: run backward smooth for a specific start and end
                           date
        _predictInSample: predict the latent state and observation for a given
                          period of time (deprecated)
        _oneDayAheadPredict: predict one day a head.
        _continuePredict: continue predicting one day after _oneDayAheadPredict
        _resetModelStatus: reset the model status to its prior status
        _setModelStatus: set the model status to a specific date
        _defaultOptions: a class to store and set default options
        _result: a class to store the results
        _copy: copy the result from the model to the _result class
        _reverseCopy: copy the result from the _result class to the model
        _checkFeatureSize: check whether the features's n matches the data's n
        _checkComponent: check whether a component is in dlm
        _getComponent: get the component if it is in dlm
        _getLatentState: get the latent state for a given component
        _getLatentCov: get the latent covariance for a given component
        _getComponentMean: get the mean of a given component
        _getComponentVar: get the variance of a given component
        _checkPlotOptions: set the correct options according to the fit
        _checkAndGetWorkingDates: get the correct filtering dates
    """
    # define the basic members
    # initialize the result
    def __init__(self, data):

        self.data = list(data)
        self.n = len(data)
        self.result = None
        self.builder = builder()
        self.Filter = None
        self.initialized = False
        self.options = self._defaultOptions()
        self.time = None
        self._printInfo = True

    # an inner class to store all options
    class _defaultOptions:
        """ All plotting and fitting options

        """
        def __init__(self):
            self.noise = 1.0
            self.stable = True

            self.plotOriginalData = True
            self.plotFilteredData = True
            self.plotSmoothedData = True
            self.plotPredictedData = True
            self.showDataPoint = True
            self.showFittedPoint = False
            self.showConfidenceInterval = True
            self.dataColor = 'black'
            self.filteredColor = 'blue'
            self.predictedColor = 'green'
            self.smoothedColor = 'red'
            self.separatePlot = True
            self.confidence = 0.95
            self.intervalType = 'ribbon'

    # an inner class to store all results
    class _result:
        """ Class to store the results

        """
        # class level (static) variables to record all names
        records = ['filteredObs', 'predictedObs', 'smoothedObs',
                   'filteredObsVar',
                   'predictedObsVar', 'smoothedObsVar', 'noiseVar',
                   'df',
                   'filteredState', 'predictedState', 'smoothedState',
                   'filteredCov', 'predictedCov', 'smoothedCov']

        # quantites to record the result
        def __init__(self, n):

            # initialize all records to be [None] * n
            for variable in self.records:
                setattr(self, variable, [None] * n)

            # record the dates that have been filtered
            self.filteredSteps = [0, -1]
            # record the dates that have been smoothed
            self.smoothedSteps = [0, -1]
            # record the last used filterType
            self.filteredType = None
            # record the current prediction status in the form of
            # [start date, current date, [predictedObs1, predictedObs2,...]]
            self.predictStatus = None

        # extend the current record by n blocks
        def _appendResult(self, n):
            for variable in self.records:
                getattr(self, variable).extend([None] * n)

        # pop out a specific date
        def _popout(self, date):
            for variable in self.records:
                getattr(self, variable).pop(date)

    # initialize the builder
    def _initialize(self):
        """ Initialize the model: initialize builder and filter.

        """
        self.builder.initialize(noise=self.options.noise)
        # if self.options.shrink == 'auto':
        #    self.Filter = kalmanFilter(discount = self.builder.discount,
        #                           shrink = 1 - min(self.builder.discount),
        #                           shrinkageMatrix = self.builder.sysVarPrior)
        # else:
        self.Filter = kalmanFilter(discount=self.builder.discount)
        self.result = self._result(self.n)
        self.initialized = True

    # use the forward filter to filter the data
    # start: the place where the filter started
    # end: the place where the filter ended
    # save: the index for dates where the filtered results should be saved,
    #       could be 'all' or 'end'
    # isForget: indicate where the filter should use the previous state as
    #         prior or just use the prior information from builder
    def _forwardFilter(self,
                       start=0,
                       end=None,
                       save='all',
                       ForgetPrevious=False,
                       renew=False):
        """ Running forwardFilter for the data for a given start and end date

        Args:
            start: the start date
            end: the end date (default to the last day of the chain)
            save: indicate the dates of which the result needs to be saved for.
                  'all' stands for (start, end), otherwise an integer between
                  start and end.
            ForgetPrevious: indicate whether the fitering should start from
                            the prior status or the previous date that has
                            been filtered.
                            (used for rolling window filtering, see @dlm)
            renew: if true, filter will refit certain days when the
                   chain gets
                   too long to add numerical stability, the length of the chain
                   is determined by the information carried on. For example,
                   when discount = 0.9, any days that are 65 days ago together
                   only carry information 1%, so we ignore
                   these days and refit the model to aid stability.
        """
        # the default value for end
        if end is None:
            end = self.n - 1

        # to see if the ff need to run or not
        if start > end:
            return None

        # also we need to make we save consectively
#        if save == 'all' and start > self.result.filteredSteps[1] + 1:
#            raise NameError('The data before start date has yet to be
#                            filtered!')

        # for rolling window run, we need to make sure the saved
        #  date is consecutive
#        if save != 'all' and end > self.result.filteredSteps[1] + 1:
#            raise NameError('The previous date needs to be filtered'
#                           + ' for rolling window!')

        # first we need to initialize the model to the correct status
        # if the start point is 0 or we want to forget the previous result
        if start == 0 or ForgetPrevious:
            self._resetModelStatus()

        # otherwise we use the result on the previous day as the prior
        else:
            if start > self.result.filteredSteps[1] + 1:
                raise NameError('The data before start date has' +
                                ' yet to be filtered! Otherwise set' +
                                ' ForgetPrevious to be True. Check the' +
                                ' <filteredSteps> in <result> object.')
            self._setModelStatus(date=start - 1)

        # we run the forward filter sequentially
        lastRenewPoint = start  # record the last renew point
        for step in range(start, end + 1):

            # first check whether we need to update evaluation or not
            if len(self.builder.dynamicComponents) > 0 or \
               len(self.builder.automaticComponents) > 0:
                self.builder.updateEvaluation(step)

            # check if rewnew is needed
            if renew and step - lastRenewPoint > self.builder.renewTerm \
               and self.builder.renewTerm > 0.0:
                # we renew the state of the day
                self._resetModelStatus()
                for innerStep in range(step - int(self.builder.renewTerm),
                                       step):
                    self.Filter.forwardFilter(self.builder.model,
                                              self.data[innerStep])
                lastRenewPoint = step

            # then we use the updated model to filter the state
            self.Filter.forwardFilter(self.builder.model, self.data[step])

            # extract the result and record
            if save == 'all' or save == step:
                self._copy(model=self.builder.model,
                           result=self.result,
                           step=step,
                           filterType='forwardFilter')

#        self.result.filteredSteps = (0, end)

    # use the backward smooth to smooth the state
    # start: the last date of the backward filtering chain
    # days: number of days to go back from start
    def _backwardSmoother(self, start=None, days=None, ignoreFuture=False):
        """ Backward smooth over filtered results for a specific start
            and number of days

        Args:
            start: the start date
            days: number of days to be smoothed starting from start towards
                  zero
            ignoreFuture: indicate whether the smoothed should start as if the
                          future data was not observed or using the future data
                          as the initial smoothing status.
        """
        # the default start date is the most recent date
        if start is None:
            start = self.n - 1

        # the default backward days number is the total length
        if days is None:
            end = 0
        else:
            end = max(start - days + 1, 0)

        # the forwardFilter has to be run before the smoother
        if self.result.filteredSteps[1] < start:
            raise NameError('The last day has to be filtered before smoothing! \
            check the <filteredSteps> in <result> object.')

        # and we record the most recent day which does not need to be smooth
        if start == self.n - 1 or ignoreFuture is True:
            self.result.smoothedState[start] = self.result.filteredState[start]
            self.result.smoothedObs[start] = self.result.filteredObs[start]
            self.result.smoothedCov[start] = self.result.filteredCov[start]
            self.result.smoothedObsVar[start] \
                = self.result.filteredObsVar[start]
            self.builder.model.noiseVar = self.result.noiseVar[start]
            start -= 1
        else:
            self.builder.model.noiseVar = self.result.noiseVar[self.n - 1]

        # empty smoothing chain, return None
        if start < end:
            return None

        # insert the previous smoothed dates
        self.builder.model.state = self.result.smoothedState[start + 1]
        self.builder.model.sysVar = self.result.smoothedCov[start + 1]

        # we smooth the result sequantially from start - 1 to end
        dates = list(range(end, start + 1))
        dates.reverse()
        for day in dates:
            # we first update the model to be correct status before smooth
            self.builder.model.prediction.state \
                = self.result.predictedState[day + 1]
            self.builder.model.prediction.sysVar \
                = self.result.predictedCov[day + 1]

            if len(self.builder.dynamicComponents) > 0 or \
               len(self.builder.automaticComponents) > 0:
                self.builder.updateEvaluation(day)

            # then we use the backward filter to filter the result
            self.Filter.backwardSmoother(
                model=self.builder.model,
                rawState=self.result.filteredState[day],
                rawSysVar=self.result.filteredCov[day])

            # extract the result
            self._copy(model=self.builder.model,
                       result=self.result,
                       step=day,
                       filterType='backwardSmoother')

#        self.result.smoothedSteps = (end, start)

    # Forecast the result based on filtered chains
    def _predictInSample(self, date, days=1):
        """ Predict the model's status based on the model of a specific day

        Args:
            date: the date the prediction is based on
            day: number of days forward that need to be predicted.

        Returns:
            A tuple. (Predicted observation, variance of the predicted
            observation)

        """

        if date + days > self.n - 1:
            raise NameError('The range is out of sample.')

        predictedObs = [None] * days
        predictedObsVar = [None] * days
        # reset the date to the date we are interested in
        self._setModelStatus(date=date)
        self.builder.model.prediction.step = 0
        for i in range(1, days):
            # update the evaluation vector
            if len(self.builder.dynamicComponents) > 0 or \
               len(self.builder.automaticComponents) > 0:
                self.builder.updateEvaluation(date + i)

            self.Filter.predict(self.builder.model)
            predictedObs[i - 1] = self.builder.model.prediction.obs
            predictedObsVar[i - 1] = self.builder.model.prediction.obsVar

        return (predictedObs, predictedObsVar)

    # feature set contains all the features for prediction.
    # It is a dictionary with key equals to the name of the component and
    # the value as the new feature (a list). The function
    # will first use the features provided in this feature dict, if not
    # found, it will fetch the default feature from the component. If
    # it could not find feature for some component, it returns an error
    def _oneDayAheadPredict(self, date, featureDict=None):
        """ One day ahead prediction based on the date and the featureDict.

        Args:
            date: the prediction starts (based on the observation before and
                  on this date)
            featureDict: the new feature value for some dynamic components.
                         must be specified in a form of {component_name: value}
                         if the feature for some dynamic component is not
                         supplied. The algorithm will use the features from
                         the old data. (which means if the prediction is out
                         of sample, then all dynamic component must be provided
                         with the new feature value)

        Returns:
            A tuple of (predicted_mean, predicted_variance)
        """
        if date > self.n - 1:
            raise NameError('The date is beyond the data range.')

        # get the correct status of the model
        self._setModelStatus(date=date)
        self._constructEvaluationForPrediction(featureDict=featureDict,
                                               date=date + 1)
        # initialize the prediction status
        self.builder.model.prediction.step = 0

        # start predicting
        self.Filter.predict(self.builder.model)

        predictedObs = self.builder.model.prediction.obs
        predictedObsVar = self.builder.model.prediction.obsVar
        self.result.predictStatus = [date, date + 1, [predictedObs]]

        return (predictedObs, predictedObsVar)

    def _continuePredict(self, featureDict=None):
        """ Continue predicting one day after _oneDayAheadPredict or
        after _continuePredict. After using
        _oneDayAheadPredict, the user can continue predicting by using
        _continuePredict and the new featureDict.

        Args:
            featureDict: the new feature value for some dynamic components.
                         see @_oneDayAheadPredict

        Returns:
            A tuple of (predicted_mean, predicted_variance)
        """
        if self.result.predictStatus is None:
            raise NameError('_continoousPredict can only be used after ' +
                            '_oneDayAheadPredict')
        currentDate = self.result.predictStatus[1]

        # need to take care of the automaticComponents, especially the
        # auto regressive component.
        for name in self.builder.automaticComponents:
            comp = self.builder.automaticComponents[name]
            if comp.componentType != 'autoReg':
                continue
            if len(self.result.predictStatus[2]) >= comp.d:
                feature = self.result.predictStatus[2][-comp.d:]
            else:
                extra = comp.d - len(self.predictStatus[2])
                feature = self.data[(currentDate - extra + 1):
                                    (currentDate + 1)] + self.result.predictStatus[2]
            if featureDict is None:
                featureDict = {}

            featureDict[name] = feature

        self._constructEvaluationForPrediction(featureDict=featureDict,
                                               date=currentDate + 1)
        self.Filter.predict(self.builder.model)
        predictedObs = self.builder.model.prediction.obs
        predictedObsVar = self.builder.model.prediction.obsVar
        self.result.predictStatus[1] += 1
        self.result.predictStatus[2].append(predictedObs)
        return (predictedObs, predictedObsVar)

    def _constructEvaluationForPrediction(self,
                                          featureDict=None,
                                          date=None):
        """ Construct the evaluation matrix based on date and featureDict.

        Used for prediction. Features provided in the featureDict will be used
        preferrably. If the feature is not found in featureDict, the algorithm
        will seek it based on the old data and the date.

        Args:
            featureDict: a dictionary containing {dynamic_component_name: value}
                         for update the feature for the corresponding component.
            date: if a dynamic component name is not found in featureDict, the
                  algorithm is using its old feature on the given date.

        """
        if featureDict is None and date is None:
            raise NameError('FeatureDict and date cannot be None ' +
                            'at the same time.')

        # find the correct evaluation vector
        if featureDict is None:
            self.builder.updateEvaluation(date)
        else:
            for i in self.builder.dynamicComponents:
                if i in featureDict:
                    self.builder.model.evaluation[
                        0, self.builder.componentIndex[i][0]:
                        (self.builder.componentIndex[i][1] + 1)] = featureDict[i]
                else:
                    if date is None:
                        raise NameError('Both date and featureDict are ' +
                                        'not provided for component ' +
                                        i)
                    comp = self.builder.dynamicComponents[i]
                    comp.updateEvaluation(date)
                    self.builder.model.evaluation[
                        0, self.builder.componentIndex[i][0]:
                        (self.builder.componentIndex[i][1] + 1)] = comp.evaluation

            for i in self.builder.automaticComponents:
                if i in featureDict:
                    self.builder.model.evaluation[
                        0, self.builder.componentIndex[i][0]:
                        (self.builder.componentIndex[i][1] + 1)] = featureDict[i]
                else:
                    if date is None:
                        raise NameError('Both date and featureDict are ' +
                                        'not provided for component ' +
                                        i)
                    comp = self.builder.automaticComponents[i]
                    comp.updateEvaluation(date)
                    self.builder.model.evaluation[
                        0, self.builder.componentIndex[i][0]:
                        (self.builder.componentIndex[i][1] + 1)] = comp.evaluation
        self.builder.model.evaluation = matrix(self.builder.model.evaluation)

# =========================== model helper function ==========================

    # to set model to a specific date
    def _setModelStatus(self, date=0):
        """ Set the model status to a specific date (the date mush have been filtered)

        """
        if date < self.result.filteredSteps[0] or \
           date > self.result.filteredSteps[1]:
            raise NameError('The date has yet to be filtered yet. ' +
                            'Check the <filteredSteps> in <result> object.')

        self._reverseCopy(model=self.builder.model,
                          result=self.result,
                          step=date)
        if len(self.builder.dynamicComponents) > 0 or \
           len(self.builder.automaticComponents) > 0:
            self.builder.updateEvaluation(date)

    # reset model to initial status
    def _resetModelStatus(self):
        """ Reset the model to the prior status

        """
        self.builder.model.state = self.builder.statePrior
        self.builder.model.sysVar = self.builder.sysVarPrior
        self.builder.model.noiseVar = self.builder.noiseVar
        self.builder.model.df = 1
        self.builder.model.initializeObservation()

    # a function used to copy result from the model to the result
    def _copy(self, model, result, step, filterType):
        """ Copy result from the model to _result class

        """

        if filterType == 'forwardFilter':
            result.filteredObs[step] = model.obs
            result.predictedObs[step] = model.prediction.obs
            result.filteredObsVar[step] = model.obsVar
            result.predictedObsVar[step] = model.prediction.obsVar
            result.filteredState[step] = model.state
            result.predictedState[step] = model.prediction.state
            result.filteredCov[step] = model.sysVar
            result.predictedCov[step] = model.prediction.sysVar
            result.noiseVar[step] = model.noiseVar
            result.df[step] = model.df

        elif filterType == 'backwardSmoother':
            result.smoothedState[step] = model.state
            result.smoothedObs[step] = model.obs
            result.smoothedCov[step] = model.sysVar
            result.smoothedObsVar[step] = model.obsVar

    def _reverseCopy(self, model, result, step):
        """ Copy result from _result class to the model

        """

        model.obs = result.filteredObs[step]
        model.prediction.obs = result.predictedObs[step]
        model.obsVar = result.filteredObsVar[step]
        model.prediction.obsVar = result.predictedObsVar[step]
        model.state = result.filteredState[step]
        model.prediction.state = result.predictedState[step]
        model.sysVar = result.filteredCov[step]
        model.prediction.sysVar = result.predictedCov[step]
        model.noiseVar = result.noiseVar[step]
        model.df = result.df[step]

    # check if the data size matches the dynamic features
    def _checkFeatureSize(self):
        """ Check features's n matches the data's n

        """
        if len(self.builder.dynamicComponents) > 0:
            for name in self.builder.dynamicComponents:
                if self.builder.dynamicComponents[name].n != self.n:
                    raise NameError('The data size of dlm and '
                                    + name + ' does not match')

    def _1DmatrixToArray(self, arrayOf1dMatrix):
        """ Change an array of 1 x 1 matrix to normal array.

        """
        return [item.tolist()[0][0] for item in arrayOf1dMatrix]

    # function to turn off printing system info
    def _printSystemInfo(self, yes):
        """ Whether the systematic infor should be printed.

        """
        if yes:
            self._printInfo = True
            self.builder._printInfo = True
        else:
            self._printInfo = False
            self.builder._printInfo = False

    def _clean(self):
        """ Clean up everything that needs to be reset. A common
        place to take care of of them.

        """
        self.result.predictStatus = None

    # function to judge whether a component is in the model
    def _checkComponent(self, name):
        """ Check whether a component is contained by the dlm.

        Args:
            name: the name of the component

        Returns:
            True or error.
        """
        if name in self.builder.staticComponents or \
           name in self.builder.dynamicComponents or \
           name in self.builder.automaticComponents:
            return True
        else:
            raise NameError('No such component.')

    # function to fetch a component
    def _fetchComponent(self, name):
        """ Get the component if the componeng is in the dlm

        Args:
            name: the name of the component

        Returns:
            The component or error.
        """
        if name in self.builder.staticComponents:
            comp = self.builder.staticComponents[name]
        elif name in self.builder.dynamicComponents:
            comp = self.builder.dynamicComponents[name]
        elif name in self.builder.automaticComponents:
            comp = self.builder.automaticComponents[name]
        else:
            raise NameError('No such component.')
        return comp

    # function to get the corresponding latent state
    def _getLatentState(self, name, filterType, start, end):
        """ Get the latent states of a given component.

        Args:
            name: the name of the component.
            filterType: the type of the latent states to be returned.
                        could be "forwardFilter", "backwardSmoother" or
                        "predict".
            start: the start date for the latent states to be returned.
            end: the end date to be returned.

        Returns:
            A list of latent states.
        """
        end += 1
        indx = self.builder.componentIndex[name]
        patten = lambda x: x if x is None else x[indx[0]:(indx[1] + 1), 0]

        if filterType == 'forwardFilter':
            return map(patten, self.result.filteredState[start:end])
        elif filterType == 'backwardSmoother':
            return map(patten, self.result.smoothedState[start:end])
        elif filterType == 'predict':
            return map(patten, self.result.predictedState[start:end])
        else:
            raise NameError('Incorrect filter type')

    # function to get the corresponding latent covariance
    def _getLatentCov(self, name, filterType, start, end):
        """ Get the latent covariance of a given component.

        Args:
            name: the name of the component.
            filterType: the type of the latent covariance to be returned.
                        could be "forwardFilter", "backwardSmoother" or
                        "predict".
            start: the start date for the latent covariance to be returned.
            end: the end date to be returned.

        Returns:
            A list of latent covariance.
        """
        end += 1
        indx = self.builder.componentIndex[name]
        patten = lambda x: x if x is None \
                 else x[indx[0]:(indx[1] + 1), indx[0]:(indx[1] + 1)]

        if filterType == 'forwardFilter':
            return map(patten, self.result.filteredCov[start:end])
        elif filterType == 'backwardSmoother':
            return map(patten, self.result.smoothedCov[start:end])
        elif filterType == 'predict':
            return map(patten, self.result.predictedCov[start:end])
        else:
            raise NameError('Incorrect filter type')

    # function to get the component mean
    def _getComponentMean(self, name, filterType, start, end):
        """ Get the mean of a given component.

        Args:
            name: the name of the component.
            filterType: the type of the mean to be returned.
                        could be "forwardFilter", "backwardSmoother" or
                        "predict".
            start: the start date for the mean to be returned.
            end: the end date to be returned.

        Returns:
            A list of mean.
        """
        end += 1
        comp = self._fetchComponent(name)
        componentState = self._getLatentState(name=name,
                                              filterType=filterType,
                                              start=start, end=end)
        result = []
        for k, i in enumerate(range(start, end)):
            if name not in self.builder.staticComponents:
                comp.updateEvaluation(i)
            result.append(dot(comp.evaluation,
                              componentState[k]).tolist()[0][0])
        return result

    # function to get the component variance
    def _getComponentVar(self, name, filterType, start, end):
        """ Get the variance of a given component.

        Args:
            name: the name of the component.
            filterType: the type of the variance to be returned.
                        could be "forwardFilter", "backwardSmoother" or
                        "predict".
            start: the start date for the variance to be returned.
            end: the end date to be returned.

        Returns:
            A list of variance.
        """
        end += 1
        comp = self._fetchComponent(name)
        componentCov = self._getLatentCov(name=name,
                                          filterType=filterType,
                                          start=start, end=end)
        result = []
        for k, i in enumerate(range(start, end)):
            if name not in self.builder.staticComponents:
                comp.updateEvaluation(i)
            result.append(dot(
                dot(comp.evaluation,
                    componentCov[k]), comp.evaluation.T).tolist()[0][0])
        return result

    # check start and end dates that has been filtered on
    def _checkAndGetWorkingDates(self, filterType):
        """ Check the filter status and return the dates that have
        been filtered according to the filterType.

        Args:
            filterType: the type of the filter to check. Could be
                       "forwardFilter", "backwardSmoother" and "predict"

        Returns:
            a tuple of (start_date, end_date)
        """
        if filterType == 'forwardFilter' or filterType == 'predict':
            if self.result.filteredSteps != [0, self.n - 1] \
               and self._printInfo:
                print('The fitlered dates are from ' +
                      str(self.result.filteredSteps[0]) +
                      ' to ' + str(self.result.filteredSteps[1]))
            start = self.result.filteredSteps[0]
            end = self.result.filteredSteps[1]

        elif filterType == 'backwardSmoother':
            if self.result.smoothedSteps != [0, self.n - 1] \
               and self._printInfo:
                print('The smoothed dates are from ' +
                      str(self.result.smoothedSteps[0]) +
                      ' to ' + str(self.result.smoothedSteps[1]))
            start = self.result.smoothedSteps[0]
            end = self.result.smoothedSteps[1]

        else:
            raise NameError('Incorrect filter type.')

        return (start, end)

    # check the filter status and automatic turn off some plots
    def _checkPlotOptions(self):
        """ Check the filter status and determine the plot options.

        """
        # adjust the option according to the filtering status
        if self.result.filteredSteps[1] == -1:
            self.options.plotFilteredData = False
            self.options.plotPredictedData = False

        if self.result.smoothedSteps[1] == -1:
            self.options.plotSmoothedData = False



import pydlm
import unittest

class testSeasonality(unittest.TestCase):

    def setUp(self):
        self.DEGREE = 7
        
    def testInitialization(self):
        newSeasonality = pydlm.modeler.seasonality.seasonality(self.DEGREE)
        newSeasonality.checkDimensions()

unittest.main()
        



import numpy as np
import unittest

from pydlm.modeler.trends import trend
from pydlm.modeler.seasonality import seasonality
from pydlm.modeler.dynamic import dynamic
from pydlm.modeler.autoReg import autoReg
from pydlm.func._dlm import _dlm


class test_dlm(unittest.TestCase):

    def setUp(self):
        self.data = [0] * 9 + [1] + [0] * 10
        self.dlm1 = _dlm(self.data)
        self.dlm2 = _dlm(self.data)
        self.dlm3 = _dlm([-1, 1, -1, 1, -1, 1, -1, 1, -1, 1, -1, 1])
        self.dlm4 = _dlm([0, 0, 0, 0, 0, 1, 1, 1, 1, 1])
        self.dlm5 = _dlm(range(100))
        self.dlm1.builder + trend(degree=1, discount=1)
        self.dlm2.builder + trend(degree=1, discount=1e-12)
        self.dlm3.builder + seasonality(period=2, discount=1)
        self.dlm4.builder + dynamic(features=[[0] for i in range(5)] +
                                    [[1] for i in range(5)], discount=1)
        self.dlm5.builder + trend(degree=1, discount=1) + \
            autoReg(degree=1, data=range(100), discount=1)
        self.dlm1._initialize()
        self.dlm2._initialize()
        self.dlm3._initialize()
        self.dlm4._initialize()
        self.dlm5._initialize()

    def testForwardFilter(self):
        self.dlm1._forwardFilter(start=0, end=19, renew=False)
        self.assertAlmostEqual(np.sum(self.dlm1.result.filteredObs[0:9]), 0)
        self.assertAlmostEqual(self.dlm1.result.filteredObs[9], 1.0/11)
        self.assertAlmostEqual(self.dlm1.result.filteredObs[19], 1.0/21)

        self.dlm2._forwardFilter(start=0, end=19)
        self.assertAlmostEqual(np.sum(self.dlm2.result.filteredObs[0:9]), 0.0)
        self.assertAlmostEqual(self.dlm2.result.filteredObs[9], 1.0)
        self.assertAlmostEqual(self.dlm2.result.filteredObs[19], 0.0)

    def testResetModelStatus(self):
        self.dlm1._forwardFilter(start = 0, end = 19, renew = False)
        self.dlm1.result.filteredSteps = (0, 19)
        self.assertAlmostEqual(self.dlm1.builder.model.obs, \
                               self.dlm1.result.filteredObs[19])

        self.dlm1._resetModelStatus()
        self.assertAlmostEqual(np.sum(self.dlm1.builder.model.state \
                                      - self.dlm1.builder.statePrior), 0.0)
    def testSetModelStatus(self):
        self.dlm1._forwardFilter(start = 0, end = 19, renew = False)
        self.dlm1.result.filteredSteps = (0, 19)
        self.assertAlmostEqual(self.dlm1.builder.model.obs, \
                               self.dlm1.result.filteredObs[19])
        self.dlm1._setModelStatus(date = 12)
        self.assertAlmostEqual(self.dlm1.builder.model.obs, \
                               self.dlm1.result.filteredObs[12])

    def testForwaredFilterConsectiveness(self):
        self.dlm1._forwardFilter(start = 0, end = 19, renew = False)
        filtered1 = self.dlm1.result.filteredObs

        self.dlm1._initialize()

        self.dlm1._forwardFilter(start = 0, end = 13)
        self.dlm1.result.filteredSteps = (0, 13)
        self.dlm1._forwardFilter(start = 13, end = 19)
        filtered2 = self.dlm1.result.filteredObs

        self.assertAlmostEqual(np.sum(np.array(filtered1) - np.array(filtered2)), 0.0)

    def testBackwardSmoother(self):
        self.dlm1._forwardFilter(start = 0, end = 19, renew = False)
        self.dlm1.result.filteredSteps = (0, 19)
        self.dlm1._backwardSmoother(start = 19)
        self.assertAlmostEqual(self.dlm1.result.smoothedObs[0], 1.0/21)
        self.assertAlmostEqual(self.dlm1.result.smoothedObs[19], 1.0/21)

        self.dlm2._forwardFilter(start = 0, end = 19)
        self.dlm2.result.filteredSteps = (0, 19)
        self.dlm2._backwardSmoother(start = 19)
        self.assertAlmostEqual(self.dlm2.result.smoothedObs[0], 0.0)
        self.assertAlmostEqual(self.dlm2.result.smoothedObs[19], 0.0)
        self.assertAlmostEqual(self.dlm2.result.smoothedObs[9], 1.0)

    def testOneDayAheadPredictWithoutDynamic(self):
        self.dlm3._forwardFilter(start=0, end=11, renew=False)
        self.dlm3.result.filteredSteps = (0, 11)
        (obs, var) = self.dlm3._oneDayAheadPredict(date=11)
        self.assertAlmostEqual(obs, -6.0/7)
        self.assertAlmostEqual(self.dlm3.result.predictStatus,
                               [11, 12, [-6.0/7]])

        (obs, var) = self.dlm3._oneDayAheadPredict(date=2)
        self.assertAlmostEqual(obs, 3.0/5)
        # notice that the two latent states always sum up to 0
        self.assertAlmostEqual(self.dlm3.result.predictStatus,
                               [2, 3, [3.0/5]])

    def testOneDayAheadPredictWithDynamic(self):
        self.dlm4._forwardFilter(start=0, end=9, renew=False)
        self.dlm4.result.filteredSteps = (0, 9)
        featureDict = {'dynamic': 2.0}
        (obs, var) = self.dlm4._oneDayAheadPredict(date=9,
                                                   featureDict=featureDict)
        self.assertAlmostEqual(obs, 5.0/6 * 2)

    def testContinuePredictWithoutDynamic(self):
        self.dlm3._forwardFilter(start=0, end=11, renew=False)
        self.dlm3.result.filteredSteps = (0, 11)
        (obs, var) = self.dlm3._oneDayAheadPredict(date=11)
        self.assertAlmostEqual(self.dlm3.result.predictStatus,
                               [11, 12, [-6.0/7]])
        (obs, var) = self.dlm3._continuePredict()
        self.assertAlmostEqual(self.dlm3.result.predictStatus,
                               [11, 13, [-6.0/7, 6.0/7]])

    def testContinuePredictWithDynamic(self):
        self.dlm4._forwardFilter(start=0, end=9, renew=False)
        self.dlm4.result.filteredSteps = (0, 9)
        featureDict = {'dynamic': 2.0}
        (obs, var) = self.dlm4._oneDayAheadPredict(date=9,
                                                   featureDict=featureDict)
        self.assertAlmostEqual(self.dlm4.result.predictStatus,
                               [9, 10, [5.0/6 * 2]])

        featureDict = {'dynamic': 3.0}
        (obs, var) = self.dlm4._continuePredict(featureDict=featureDict)
        self.assertAlmostEqual(self.dlm4.result.predictStatus,
                               [9, 11, [5.0/6 * 2, 5.0/6 * 3]])

    def testPredictWithAutoReg(self):
        self.dlm5._forwardFilter(start=0, end=99, renew=False)
        self.dlm5.result.filteredSteps = [0, 99]
        (obs, var) = self.dlm5._oneDayAheadPredict(date=99)
        self.assertAlmostEqual(obs, 100.03682874)
        (obs, var) = self.dlm5._continuePredict()
        self.assertAlmostEqual(obs, 101.07480945)

    def testGetLatentState(self):
        # for forward filter
        self.dlm5._forwardFilter(start=0, end=99, renew=False)
        self.dlm5.result.filteredSteps = [0, 99]
        filteredTrend = self.dlm5._getLatentState(
            filterType='forwardFilter', name='trend', start=0, end=99)
        diff = 0.0
        for i in range(len(filteredTrend)):
            diff += abs(filteredTrend[i][0] -
                        self.dlm5.result.filteredState[i][0, 0])
        self.assertAlmostEqual(diff, 0)

        # for prediction
        predictedTrend = self.dlm5._getLatentState(
            filterType='predict', name='trend', start=0, end=99)
        diff = 0.0
        for i in range(len(predictedTrend)):
            diff += abs(predictedTrend[i][0] -
                        self.dlm5.result.predictedState[i][0, 0])
        self.assertAlmostEqual(diff, 0)

        # for backward smoother
        self.dlm5._backwardSmoother(start=99)
        self.dlm5.result.smoothedSteps = [0, 99]
        smoothedTrend = self.dlm5._getLatentState(
            filterType='backwardSmoother', name='trend', start=0, end=99)
        diff = 0.0
        for i in range(len(smoothedTrend)):
            diff += abs(smoothedTrend[i][0] -
                        self.dlm5.result.smoothedState[i][0, 0])
        self.assertAlmostEqual(diff, 0)

    def testGetLatentCov(self):
        # for forward filter
        self.dlm5._forwardFilter(start=0, end=99, renew=False)
        self.dlm5.result.filteredSteps = [0, 99]
        filteredTrend = self.dlm5._getLatentCov(
            filterType='forwardFilter', name='trend', start=0, end=99)
        diff = 0.0
        for i in range(len(filteredTrend)):
            diff += abs(filteredTrend[i][0, 0] -
                        self.dlm5.result.filteredCov[i][0, 0])
        self.assertAlmostEqual(diff, 0)

        # for prediction
        predictedTrend = self.dlm5._getLatentCov(
            filterType='predict', name='trend', start=0, end=99)
        diff = 0.0
        for i in range(len(predictedTrend)):
            diff += abs(predictedTrend[i][0, 0] -
                        self.dlm5.result.predictedCov[i][0, 0])
        self.assertAlmostEqual(diff, 0)

        # for backward smoother
        self.dlm5._backwardSmoother(start=99)
        self.dlm5.result.smoothedSteps = [0, 99]
        smoothedTrend = self.dlm5._getLatentCov(
            filterType='backwardSmoother', name='trend', start=0, end=99)
        diff = 0.0
        for i in range(len(smoothedTrend)):
            diff += abs(smoothedTrend[i][0, 0] -
                        self.dlm5.result.smoothedCov[i][0, 0])
        self.assertAlmostEqual(diff, 0)

    def testComponentMean(self):
        self.dlm5._forwardFilter(start=0, end=99, renew=False)
        self.dlm5.result.filteredSteps = [0, 99]
        # for component with forward filter
        arTrend = self.dlm5._getComponentMean(filterType='forwardFilter',
                                              name='ar2', start=0, end=99)
        trueAr = [item[1, 0] for item in self.dlm5.result.filteredState]
        comp = self.dlm5.builder.automaticComponents['ar2']
        for i in range(len(trueAr)):
            comp.updateEvaluation(i)
            trueAr[i] = comp.evaluation * trueAr[i]

        diff = 0.0
        for i in range(len(arTrend)):
            diff += abs(arTrend[i] - trueAr[i])
        self.assertAlmostEqual(diff, 0)

        # for component with backward smoother
        self.dlm5._backwardSmoother(start=99)
        self.dlm5.result.smoothedSteps = [0, 99]
        arTrend = self.dlm5._getComponentMean(filterType='backwardSmoother',
                                    name='ar2', start=0, end=99)
        trueAr = [item[1, 0] for item in self.dlm5.result.smoothedState]
        comp = self.dlm5.builder.automaticComponents['ar2']
        for i in range(len(trueAr)):
            comp.updateEvaluation(i)
            trueAr[i] = comp.evaluation * trueAr[i]

        diff = 0.0
        for i in range(len(arTrend)):
            diff += abs(arTrend[i] - trueAr[i])
        self.assertAlmostEqual(diff, 0)

    def testComponentVar(self):
        self.dlm5._forwardFilter(start=0, end=99, renew=False)
        self.dlm5.result.filteredSteps = [0, 99]
        # for component with forward filter
        arTrend = self.dlm5._getComponentVar(filterType='forwardFilter',
                                             name='ar2', start=0, end=99)
        trueAr = [item[1, 1] for item in self.dlm5.result.filteredCov]
        comp = self.dlm5.builder.automaticComponents['ar2']
        for i in range(len(trueAr)):
            comp.updateEvaluation(i)
            trueAr[i] = comp.evaluation * trueAr[i] * comp.evaluation.T

        diff = 0.0
        for i in range(len(arTrend)):
            diff += abs(arTrend[i] - trueAr[i])
        self.assertAlmostEqual(diff, 0)

        # for component with backward smoother
        self.dlm5._backwardSmoother(start=99)
        self.dlm5.result.smoothedSteps = [0, 99]
        arTrend = self.dlm5._getComponentVar(filterType='backwardSmoother',
                                             name='ar2', start=0, end=99)
        trueAr = [item[1, 1] for item in self.dlm5.result.smoothedCov]
        comp = self.dlm5.builder.automaticComponents['ar2']
        for i in range(len(trueAr)):
            comp.updateEvaluation(i)
            trueAr[i] = comp.evaluation * trueAr[i] * comp.evaluation.T

        diff = 0.0
        for i in range(len(arTrend)):
            diff += abs(arTrend[i] - trueAr[i])
        self.assertAlmostEqual(diff, 0)

unittest.main()



import numpy as np
import unittest

from pydlm.modeler.trends import trend
from pydlm.modeler.seasonality import seasonality
from pydlm.modeler.dynamic import dynamic
from pydlm.modeler.autoReg import autoReg
from pydlm.dlm import dlm

class testDlm(unittest.TestCase):

    def setUp(self):
        self.data = [0] * 9 + [1] + [0] * 10
        self.features = np.random.random((20, 2)).tolist()
        self.dlm1 = dlm(self.data)
        self.dlm2 = dlm(self.data)
        self.dlm3 = dlm([-1, 1, -1, 1, -1, 1, -1, 1, -1, 1, -1, 1])
        self.dlm4 = dlm([0, 0, 0, 0, 0, 1, 1, 1, 1, 1])
        self.dlm5 = dlm(range(100))
        self.dlm1 + trend(degree = 1, discount = 1)
        self.dlm2 + trend(degree = 1, discount = 1e-12)
        self.dlm3 + seasonality(period = 2, discount = 1)
        self.dlm4 + dynamic(features=[[0] for i in range(5)] +
                            [[1] for i in range(5)], discount=1)
        self.dlm5 + trend(degree=1, discount=1) + \
            autoReg(degree=1, data=range(100), discount=1)

    def testAdd(self):
        trend2 = trend(2, name='trend2')
        self.dlm1 = self.dlm1 + trend2
        self.assertEqual(self.dlm1.builder.staticComponents['trend2'], trend2)

        dynamic2 = dynamic(features=self.features, name='d2')
        self.dlm1 = self.dlm1 + dynamic2
        self.assertEqual(self.dlm1.builder.dynamicComponents['d2'], dynamic2)

        ar3 = autoReg(degree=3,data=self.data, name='ar3')
        self.dlm1 = self.dlm1 + ar3
        self.assertEqual(self.dlm1.builder.automaticComponents['ar3'], ar3)

    def testDelete(self):
        trend2 = trend(2, name='trend2')
        self.dlm1 = self.dlm1 + trend2
        self.dlm1.delete('trend2')
        self.assertEqual(len(self.dlm1.builder.staticComponents), 1)

    def testFitForwardFilter(self):
        self.dlm1.fitForwardFilter(useRollingWindow = False)
        self.assertEqual(self.dlm1.result.filteredSteps, [0, 19])
        self.assertAlmostEqual(np.sum(self.dlm1.result.filteredObs[0:9]), 0)
        self.assertAlmostEqual(self.dlm1.result.filteredObs[9], 1.0/11)
        self.assertAlmostEqual(self.dlm1.result.filteredObs[19], 1.0/21)

        self.dlm2.fitForwardFilter(useRollingWindow = False)
        self.assertAlmostEqual(np.sum(self.dlm2.result.filteredObs[0:9]), 0.0)
        self.assertAlmostEqual(self.dlm2.result.filteredObs[9], 1.0)
        self.assertAlmostEqual(self.dlm2.result.filteredObs[19], 0.0)

    def testFitBackwardSmoother(self):
        self.dlm1.fitForwardFilter()
        self.dlm1.fitBackwardSmoother()
        self.assertEqual(self.dlm1.result.smoothedSteps, [0, 19])
        self.assertAlmostEqual(self.dlm1.result.smoothedObs[0], 1.0/21)
        self.assertAlmostEqual(self.dlm1.result.smoothedObs[19], 1.0/21)

        self.dlm2.fitForwardFilter()
        self.dlm2.fitBackwardSmoother()
        self.assertAlmostEqual(self.dlm2.result.smoothedObs[0], 0.0)
        self.assertAlmostEqual(self.dlm2.result.smoothedObs[19], 0.0)
        self.assertAlmostEqual(self.dlm2.result.smoothedObs[9], 1.0)

    def testAppend(self):
        dlm4 = dlm(self.data[0:11])
        dlm4 + trend(degree = 1, discount = 1)
        dlm4.fitForwardFilter()
        self.assertEqual(dlm4.n, 11)

        dlm4.append(self.data[11 : 20])
        self.assertEqual(dlm4.n, 20)
        dlm4.fitForwardFilter()

        self.dlm1.fitForwardFilter()
        self.assertAlmostEqual(np.sum(np.array(dlm4.result.filteredObs) - \
                                      np.array(self.dlm1.result.filteredObs)), 0.0)

    def testAppendDynamic(self):
        # we feed the data to dlm4 via two segments
        dlm4 = dlm(self.data[0:11])
        dlm4 + trend(degree = 1, discount = 1) + dynamic(features = self.features[0:11], \
                                                         discount = 1)
        dlm4.fitForwardFilter()
        dlm4.append(self.data[11 : 20])
        dlm4.append(self.features[11 : 20], component = 'dynamic')
        dlm4.fitForwardFilter()

        # we feed the data to dlm5 all at once
        dlm5 = dlm(self.data)
        dlm5 + trend(degree = 1, discount = 1) + dynamic(features = self.features, \
                                                         discount = 1)
        dlm5.fitForwardFilter()
        self.assertAlmostEqual(np.sum(np.array(dlm4.result.filteredObs) - \
                                      np.array(dlm5.result.filteredObs)), 0.0)

    def testAppendAutomatic(self):
        # we feed the data to dlm4 via two segments
        dlm4 = dlm(self.data[0:11])
        dlm4 + trend(degree = 1, discount = 1) + autoReg(degree = 3,
                                                         data = self.data[0:11],
                                                         discount = 1)
        dlm4.fitForwardFilter()
        dlm4.append(self.data[11 : 20])
        dlm4.fitForwardFilter()

        # we feed the data to dlm5 all at once
        dlm5 = dlm(self.data)
        dlm5 + trend(degree = 1, discount = 1) + autoReg(degree = 3,
                                                         data = self.data,
                                                         discount = 1)
        dlm5.fitForwardFilter()
        self.assertAlmostEqual(np.sum(np.array(dlm4.result.filteredObs) - \
                                      np.array(dlm5.result.filteredObs)), 0.0)

    def testPopout(self):
        dlm4 = dlm(self.data)
        dlm4 + trend(degree = 1, discount = 1) + dynamic(features = \
                                                         self.features, \
                                                         discount = 1)
        dlm4.fitForwardFilter()
        # the filtered step range should be (0, 19)
        self.assertEqual(dlm4.result.filteredSteps, [0, 19])

        # pop out the first date, the filtered range should be (0, -1)
        dlm4.popout(0)
        self.assertEqual(dlm4.result.filteredSteps, [0, -1])

        dlm4.fitForwardFilter()
        dlm5 = dlm(self.data[1 : 20])
        dlm5 + trend(degree = 1, discount = 1) + dynamic(features = \
                                                         self.features[1 : 20], \
                                                         discount = 1)
        dlm5.fitForwardFilter()

        # The two chain should have the same filtered obs
        self.assertAlmostEqual(np.sum(np.array(dlm4.result.filteredObs) - \
                                      np.array(dlm5.result.filteredObs)), 0.0)

    def testAlter(self):
        dlm4 = dlm(self.data)
        dlm4 + trend(degree = 1, discount = 1) + dynamic(features = \
                                                         self.features, \
                                                         discount = 1)
        dlm4.fitForwardFilter()
        # the filtered step range should be (0, 19)
        self.assertEqual(dlm4.result.filteredSteps, [0, 19])

        # pop out the first date, the filtered range should be (0, -1)
        dlm4.alter(date = 15, data = 1, component = 'main')
        self.assertEqual(dlm4.result.filteredSteps, [0, 14])
        dlm4.fitForwardFilter()

        newData = [0] * 9 + [1] + [0] * 10
        newData[15] = 1
        dlm5 = dlm(newData)
        dlm5 + trend(degree = 1, discount = 1) + dynamic(features = \
                                                         self.features, \
                                                         discount = 1)
        dlm5.fitForwardFilter()

        # The two chain should have the same filtered obs
        self.assertAlmostEqual(np.sum(np.array(dlm4.result.filteredObs) - \
                                      np.array(dlm5.result.filteredObs)), 0.0)

    def testOneDayAheadPredictWithoutDynamic(self):
        self.dlm3.fitForwardFilter()
        (obs, var) = self.dlm3.predict(date=11)
        self.assertAlmostEqual(obs, -6.0/7)
        self.assertAlmostEqual(self.dlm3.result.predictStatus,
                               [11, 12, [-6.0/7]])

        (obs, var) = self.dlm3.predict(date=2)
        self.assertAlmostEqual(obs, 3.0/5)
        # notice that the two latent states always sum up to 0
        self.assertAlmostEqual(self.dlm3.result.predictStatus,
                               [2, 3, [3.0/5]])

    def testOneDayAheadPredictWithDynamic(self):
        self.dlm4.fitForwardFilter()
        featureDict = {'dynamic': 2.0}
        (obs, var) = self.dlm4.predict(date=9,
                                       featureDict=featureDict)
        self.assertAlmostEqual(obs, 5.0/6 * 2)

    def testContinuePredictWithoutDynamic(self):
        self.dlm3.fitForwardFilter()
        (obs, var) = self.dlm3.predict(date=11)
        self.assertAlmostEqual(self.dlm3.result.predictStatus,
                               [11, 12, [-6.0/7]])
        (obs, var) = self.dlm3.continuePredict()
        self.assertAlmostEqual(self.dlm3.result.predictStatus,
                               [11, 13, [-6.0/7, 6.0/7]])

    def testContinuePredictWithDynamic(self):
        self.dlm4.fitForwardFilter()
        featureDict = {'dynamic': 2.0}
        (obs, var) = self.dlm4.predict(date=9,
                                       featureDict=featureDict)
        self.assertAlmostEqual(self.dlm4.result.predictStatus,
                               [9, 10, [5.0/6 * 2]])

        featureDict = {'dynamic': 3.0}
        (obs, var) = self.dlm4.continuePredict(featureDict=featureDict)
        self.assertAlmostEqual(self.dlm4.result.predictStatus,
                               [9, 11, [5.0/6 * 2, 5.0/6 * 3]])

    def testPredictWithAutoReg(self):
        self.dlm5.fitForwardFilter()
        (obs, var) = self.dlm5.predict(date=99)
        self.assertAlmostEqual(obs, 100.03682874)
        (obs, var) = self.dlm5.continuePredict()
        self.assertAlmostEqual(obs, 101.07480945)

    def testGetLatentState(self):
        # for forward filter
        self.dlm5.fitForwardFilter()
        filteredTrend = self.dlm5.getLatentState(
            filterType='forwardFilter', name='trend')
        diff = 0.0
        for i in range(len(filteredTrend)):
            diff += abs(filteredTrend[i][0] -
                        self.dlm5.result.filteredState[i][0, 0])
        self.assertAlmostEqual(diff, 0)

        # for backward smoother
        self.dlm5.fitBackwardSmoother()
        smoothedTrend = self.dlm5.getLatentState(
            filterType='backwardSmoother', name='trend')
        diff = 0.0
        for i in range(len(smoothedTrend)):
            diff += abs(smoothedTrend[i][0] -
                        self.dlm5.result.smoothedState[i][0, 0])
        self.assertAlmostEqual(diff, 0)

    def testGetLatentCov(self):
        # for forward filter
        self.dlm5.fitForwardFilter()
        filteredTrend = self.dlm5.getLatentCov(
            filterType='forwardFilter', name='trend')
        diff = 0.0
        for i in range(len(filteredTrend)):
            diff += abs(filteredTrend[i][0, 0] -
                        self.dlm5.result.filteredCov[i][0, 0])
        self.assertAlmostEqual(diff, 0)

        # for backward smoother
        self.dlm5.fitBackwardSmoother()
        smoothedTrend = self.dlm5.getLatentCov(
            filterType='backwardSmoother', name='trend')
        diff = 0.0
        for i in range(len(smoothedTrend)):
            diff += abs(smoothedTrend[i][0, 0] -
                        self.dlm5.result.smoothedCov[i][0, 0])
        self.assertAlmostEqual(diff, 0)

    def testGetMean(self):
        # for forward filter
        self.dlm5.fitForwardFilter()
        filteredTrend = self.dlm5.getMean(filterType='forwardFilter')
        diff = 0.0
        for i in range(len(filteredTrend)):
            diff += abs(filteredTrend[i] -
                        self.dlm5.result.filteredObs[i][0, 0])
        self.assertAlmostEqual(diff, 0)

        # for component with forward filter
        arTrend = self.dlm5.getMean(filterType='forwardFilter',
                                    name='ar2')
        trueAr = [item[1, 0] for item in self.dlm5.result.filteredState]
        comp = self.dlm5.builder.automaticComponents['ar2']
        for i in range(len(trueAr)):
            comp.updateEvaluation(i)
            trueAr[i] = comp.evaluation * trueAr[i]

        diff = 0.0
        for i in range(len(arTrend)):
            diff += abs(arTrend[i] - trueAr[i])
        self.assertAlmostEqual(diff, 0)

        # for backward smoother
        self.dlm5.fitBackwardSmoother()
        filteredTrend = self.dlm5.getMean(filterType='backwardSmoother')
        diff = 0.0
        for i in range(len(filteredTrend)):
            diff += abs(filteredTrend[i] -
                        self.dlm5.result.smoothedObs[i][0, 0])
        self.assertAlmostEqual(diff, 0)

        # for component with backward smoother
        arTrend = self.dlm5.getMean(filterType='backwardSmoother',
                                    name='ar2')
        trueAr = [item[1, 0] for item in self.dlm5.result.smoothedState]
        comp = self.dlm5.builder.automaticComponents['ar2']
        for i in range(len(trueAr)):
            comp.updateEvaluation(i)
            trueAr[i] = comp.evaluation * trueAr[i]

        diff = 0.0
        for i in range(len(arTrend)):
            diff += abs(arTrend[i] - trueAr[i])
        self.assertAlmostEqual(diff, 0)

    def testGetVar(self):
        # for forward filter
        self.dlm5.fitForwardFilter()
        filteredTrend = self.dlm5.getVar(filterType='forwardFilter')
        diff = 0.0
        for i in range(len(filteredTrend)):
            diff += abs(filteredTrend[i] -
                        self.dlm5.result.filteredObsVar[i][0, 0])
        self.assertAlmostEqual(diff, 0)

        # for component with forward filter
        arTrend = self.dlm5.getVar(filterType='forwardFilter',
                                    name='ar2')
        trueAr = [item[1, 1] for item in self.dlm5.result.filteredCov]
        comp = self.dlm5.builder.automaticComponents['ar2']
        for i in range(len(trueAr)):
            comp.updateEvaluation(i)
            trueAr[i] = comp.evaluation * trueAr[i] * comp.evaluation.T

        diff = 0.0
        for i in range(len(arTrend)):
            diff += abs(arTrend[i] - trueAr[i])
        self.assertAlmostEqual(diff, 0)

        # for backward smoother
        self.dlm5.fitBackwardSmoother()
        filteredTrend = self.dlm5.getVar(filterType='backwardSmoother')
        diff = 0.0
        for i in range(len(filteredTrend)):
            diff += abs(filteredTrend[i] -
                        self.dlm5.result.smoothedObsVar[i][0, 0])
        self.assertAlmostEqual(diff, 0)

        # for component with backward smoother
        arTrend = self.dlm5.getVar(filterType='backwardSmoother',
                                    name='ar2')
        trueAr = [item[1, 1] for item in self.dlm5.result.smoothedCov]
        comp = self.dlm5.builder.automaticComponents['ar2']
        for i in range(len(trueAr)):
            comp.updateEvaluation(i)
            trueAr[i] = comp.evaluation * trueAr[i] * comp.evaluation.T

        diff = 0.0
        for i in range(len(arTrend)):
            diff += abs(arTrend[i] - trueAr[i])
        self.assertAlmostEqual(diff, 0)

unittest.main()



#import numpy as np
#import pydlm as pd

#data = np.concatenate((np.random.random(100), np.random.random(100) + 3))
#myDLM = pd.dlm(data) + pd.trend(2, discount = 0.9)


#myDLM.turnOn('smooth')
#myDLM.turnOn('predict')
#myDLM.turnOff('multiple')
#myDLM.shrink(0.0)
#myDLM.fitForwardFilter(useRollingWindow = False, windowLength = 20)
#myDLM.fitForwardFilter()
#myDLM.fitBackwardSmoother()
#myDLM.plot()
